# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Setup/teardown fixtures common to unittest/pytest.

* Unittest: Those functions are wrapped in `@contextlib.contextmanager` and
  called in `setUpClass`/`tearDownClass`.
* Pytest: Those functions are wrapped in `@fixture`.

This file is not part of `tensorflow_datasets.testing` because this module still
heavily depends on TensorFlow.
"""

from unittest import mock

from tensorflow_datasets.core import load
from tensorflow_datasets.core.github_api import github_path


def assert_no_api_call():
  """Globally disable github API calls."""
  with mock.patch.object(
      github_path.GithubApi,  # pylint: disable=protected-access
      'query',
      side_effect=AssertionError('Forbidden API call'),
  ):
    yield


# Fixtures automatically applied for all tests (unittest and pytest)
GLOBAL_FIXTURES = [
    assert_no_api_call,
]



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for beam_utils."""

import os
import pathlib
from typing import Optional

import pytest
from tensorflow_datasets.core import beam_utils
from tensorflow_datasets.core import dataset_builder
from tensorflow_datasets.core import dataset_utils


@pytest.mark.parametrize(
    'split, workers_per_shard, expected_result',
    [
        ('train', 1, "{'id': 0}\n{'id': 1}\n{'id': 2}\n"),
        ('train', 2, "{'id': 0}\n{'id': 1}\n{'id': 2}\n"),
        ('train', 100, "{'id': 0}\n{'id': 1}\n{'id': 2}\n"),
        ('train[:2]', 1, "{'id': 0}\n{'id': 1}\n"),
        ('train[1:]', 1, "{'id': 1}\n{'id': 2}\n"),
    ],
)
def test_read_from_tfds(
    dummy_dataset: dataset_builder.DatasetBuilder,
    tmp_path: pathlib.Path,
    split: str,
    workers_per_shard: Optional[int],
    expected_result: str,
):
  import apache_beam as beam  # pylint: disable=g-import-not-at-top

  with beam.Pipeline() as pipeline:
    _ = (
        pipeline
        | beam_utils.ReadFromTFDS(
            dummy_dataset, split=split, workers_per_shard=workers_per_shard
        )
        | beam.Map(dataset_utils.as_numpy)
        | beam.io.WriteToText(os.fspath(tmp_path / 'out.txt'))
    )

  assert (tmp_path / 'out.txt-00000-of-00001').read_text() == expected_result


@pytest.mark.parametrize(
    'split, expected_implemented_with_batchsize',
    [
        ('train', True),
        ('train[:2]', False),
        ('train[1:]', False),
    ],
)
def test_subsplit_failure_with_batch_size(
    dummy_dataset: dataset_builder.DatasetBuilder,
    tmp_path: pathlib.Path,
    split,
    expected_implemented_with_batchsize,
):
  implemented_with_batchsize = True
  try:
    import apache_beam as beam  # pylint: disable=g-import-not-at-top

    with beam.Pipeline() as pipeline:
      _ = (
          pipeline
          | beam_utils.ReadFromTFDS(dummy_dataset, split=split, batch_size=2)
          | beam.Map(dataset_utils.as_numpy)
          | beam.io.WriteToText(os.fspath(tmp_path / 'out.txt'))
      )

  except NotImplementedError:
    implemented_with_batchsize = False
  assert implemented_with_batchsize == expected_implemented_with_batchsize



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Public API of the proto package."""
# pylint: disable=g-import-not-at-top,g-importing-member, import-outside-toplevel

try:  # pylint: disable=g-statement-before-imports
  from waymo_open_dataset import dataset_pb2 as waymo_dataset_pb2
except ImportError:
  # If original waymo proto is not found, fallback to the pre-generated proto
  from tensorflow_datasets.proto import waymo_dataset_generated_pb2 as waymo_dataset_pb2  # pylint: disable=line-too-long

from tensorflow_datasets.proto import tf_example_generated_pb2 as tf_example_pb2
from tensorflow_datasets.proto import tf_feature_generated_pb2 as tf_feature_pb2



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities to manipulate images.

Note: these functions are not meant to be used inside of a TF graph.
"""
from __future__ import annotations

import csv
import subprocess
from typing import Any, List, Optional

import numpy as np
from tensorflow_datasets.core import lazy_imports_lib
from tensorflow_datasets.core.utils import py_utils
from tensorflow_datasets.core.utils import resource_utils
from tensorflow_datasets.core.utils import tf_utils
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf

PilImage = Any  # Require lazy deps.
THUMBNAIL_SIZE = 128


@py_utils.memoize()
def _get_runner():
  return tf_utils.TFGraphRunner()


def decode_image(image_bytes: bytes) -> np.ndarray:
  """Returns np.array corresponding to encoded image."""
  runner = _get_runner()
  return runner.run(tf.image.decode_image, image_bytes)


def png_to_jpeg(image_bytes: bytes, quality: int = 100) -> bytes:
  """Converts PNG image (bytes or str) to JPEG (bytes)."""
  runner = _get_runner()
  decode_fn = lambda img: tf.image.decode_png(img, channels=3)
  image = runner.run(decode_fn, image_bytes)
  fn = lambda img: tf.image.encode_jpeg(img, format='rgb', quality=quality)
  return runner.run(fn, image)


def jpeg_cmyk_to_rgb(image_bytes: bytes, quality: int = 100) -> bytes:
  """Converts JPEG CMYK image (bytes) to RGB JPEG (bytes)."""
  runner = _get_runner()
  image = runner.run(tf.image.decode_jpeg, image_bytes)
  fn = lambda img: tf.image.encode_jpeg(img, format='rgb', quality=quality)
  return runner.run(fn, image)


def ffmpeg_run(
    args: List[str],
    stdin: Optional[bytes] = None,
    timeout: Optional[int] = None,
) -> bytes:
  """Executes the ffmpeg function.

  Args:
    args: A list of string args and flags to send to the ffmpeg binary.
    stdin: Bytes to provide as standard input, or None.
    timeout: Maximum amount of seconds to wait for, before interrupting the
      command. None to wait forever.

  Returns:
    The stdout output.
  """
  ffmpeg_path = 'ffmpeg'
  try:
    cmd_args = [ffmpeg_path] + args
    return subprocess.run(
        cmd_args,
        check=True,
        input=stdin,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        timeout=timeout,
    ).stdout
  except subprocess.CalledProcessError as e:
    raise ValueError(
        f'Command {e.cmd} returned error code {e.returncode}:\n'
        f'stdout={e.stdout.decode("utf-8")}\n'
        f'stderr={e.stderr.decode("utf-8")}\n'
    ) from e
  except FileNotFoundError as e:
    raise FileNotFoundError(
        'It seems that ffmpeg is not installed on the system. Please follow '
        'the instrutions at https://ffmpeg.org/. '
        f'Original exception: {e}'
    ) from e


@py_utils.memoize()
def get_colormap() -> np.ndarray:
  """Loads the colormap.

  The colormap was precomputed using Glasbey et al. algorythm (Colour Displays
  for Categorical Images, 2017) to generate maximally distinct colors.

  It was generated using https://github.com/taketwo/glasbey:

  ```python
  gb = glasbey.Glasbey(
      base_palette=[(0, 0, 0), (228, 26, 28), (55, 126, 184), (77, 175, 74)],
      no_black=True,
  )
  palette = gb.generate_palette(size=256)
  gb.save_palette(palette, 'colormap.csv')
  ```

  Returns:
    colormap: A `np.array(shape=(255, 3), dtype=np.uint8)` representing the
      mapping id -> color.
  """
  colormap_path = resource_utils.tfds_path() / 'core/utils/colormap.csv'
  with colormap_path.open() as f:
    return np.array(list(csv.reader(f)), dtype=np.uint8)


def apply_colormap(image: np.ndarray) -> np.ndarray:
  """Apply colormap from grayscale (h, w, 1) to colored (h, w, 3) image."""
  image = image.squeeze(axis=-1)  # (h, w, 1) -> (h, w)
  cmap = get_colormap()  # Get the (256, 3) colormap
  # Normalize uint16 and convert each value to a unique color
  return cmap[image % len(cmap)]


# Visualization single image


def _postprocess_noop(img: PilImage) -> PilImage:
  return img


def _postprocess_convert_rgb(img: PilImage) -> PilImage:
  return img.convert('RGB')


def create_thumbnail(
    ex: np.ndarray, *, use_colormap: bool, default_dimensions: bool = True
) -> PilImage:
  """Creates the image from the np.array input."""
  PIL_Image = lazy_imports_lib.lazy_imports.PIL_Image  # pylint: disable=invalid-name

  if use_colormap:  # Apply the colormap first as it modify the shape/dtype
    ex = apply_colormap(ex)

  _, _, c = ex.shape
  postprocess = _postprocess_noop
  if c == 1:
    ex = ex.squeeze(axis=-1)
    mode = 'L'
  elif ex.dtype == np.uint16:
    mode = 'I;16'
    postprocess = _postprocess_convert_rgb
  else:
    mode = None
  img = PIL_Image.fromarray(ex, mode=mode)
  img = postprocess(img)
  if default_dimensions:
    img.thumbnail((THUMBNAIL_SIZE, THUMBNAIL_SIZE))  # Resize the image in-place
  return img



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Image Classification Folder datasets."""

from __future__ import annotations

import collections
import os
import random
from typing import Dict, List, NoReturn, Optional, Tuple

from tensorflow_datasets.core import dataset_builder
from tensorflow_datasets.core import dataset_info
from tensorflow_datasets.core import decode
from tensorflow_datasets.core import features as features_lib
from tensorflow_datasets.core import splits as split_lib
from tensorflow_datasets.core.utils import type_utils
from tensorflow_datasets.core.utils import version
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf

_SUPPORTED_IMAGE_FORMAT = ('.jpg', '.jpeg', '.png')

_Example = collections.namedtuple('_Example', ['image_path', 'label'])

# Dict of 'split_name'-> `List[_Example]`
SplitExampleDict = Dict[str, List[_Example]]


class ImageFolder(dataset_builder.DatasetBuilder):
  """Generic image classification dataset created from manual directory.

  `ImageFolder` creates a `tf.data.Dataset` reading the original image files.

  The data directory should have the following structure:

  ```
  path/to/image_dir/
    split_name/  # Ex: 'train'
      label1/  # Ex: 'airplane' or '0015'
        xxx.png
        xxy.png
        xxz.png
      label2/
        xxx.png
        xxy.png
        xxz.png
    split_name/  # Ex: 'test'
      ...
  ```

  To use it:

  ```
  builder = tfds.ImageFolder('path/to/image_dir/')
  print(builder.info)  # num examples, labels... are automatically calculated
  ds = builder.as_dataset(split='train', shuffle_files=True)
  tfds.show_examples(ds, builder.info)
  ```
  """

  VERSION = version.Version('1.0.0')

  def __init__(
      self,
      root_dir: str,
      *,
      shape: Optional[type_utils.Shape] = None,
      dtype: Optional[tf.DType] = None,
  ):
    """Construct the `DatasetBuilder`.

    Args:
      root_dir: Path to the directory containing the images.
      shape: Image shape forwarded to `tfds.features.Image`.
      dtype: Image dtype forwarded to `tfds.features.Image`.
    """
    self._image_shape = shape
    self._image_dtype = dtype
    super(ImageFolder, self).__init__()
    self._data_dir = root_dir  # Set data_dir to the existing dir.

    # Extract the splits, examples, labels
    root_dir = os.path.expanduser(root_dir)
    self._split_examples, labels = _get_split_label_images(root_dir)

    # Update DatasetInfo labels
    self.info.features['label'].names = sorted(labels)

    # Update DatasetInfo splits
    split_infos = [
        split_lib.SplitInfo(  # pylint: disable=g-complex-comprehension
            name=split_name,
            shard_lengths=[len(examples)],
            num_bytes=0,
        )
        for split_name, examples in self._split_examples.items()
    ]
    split_dict = split_lib.SplitDict(split_infos)
    self.info.set_splits(split_dict)

  def _info(self) -> dataset_info.DatasetInfo:
    return dataset_info.DatasetInfo(
        builder=self,
        description='Generic image classification dataset.',
        features=features_lib.FeaturesDict({
            'image': features_lib.Image(
                shape=self._image_shape,
                dtype=self._image_dtype,
            ),
            'label': features_lib.ClassLabel(),
            'image/filename': features_lib.Text(),
        }),
        supervised_keys=('image', 'label'),
    )

  def _download_and_prepare(self, **kwargs) -> NoReturn:  # pytype: disable=signature-mismatch  # overriding-parameter-count-checks
    raise NotImplementedError(
        'No need to call download_and_prepare function for {}.'.format(
            type(self).__name__
        )
    )

  def download_and_prepare(self, **kwargs):  # -> NoReturn:
    return self._download_and_prepare()

  def _as_dataset(
      self,
      split: str,
      shuffle_files: bool = False,
      decoders: Optional[Dict[str, decode.Decoder]] = None,
      read_config=None,
  ) -> tf.data.Dataset:
    """Generate dataset for given split."""
    del read_config  # Unused (automatically created in `DatasetBuilder`)

    if split not in self.info.splits.keys():
      raise ValueError(
          'Unrecognized split {}. Subsplit API not yet supported for {}. '
          'Split name should be one of {}.'.format(
              split, type(self).__name__, list(self.info.splits.keys())
          )
      )

    # Extract all labels/images
    image_paths = []
    labels = []
    examples = self._split_examples[split]
    for example in examples:
      image_paths.append(example.image_path)
      labels.append(self.info.features['label'].str2int(example.label))

    # Build the tf.data.Dataset object
    ds = tf.data.Dataset.from_tensor_slices((image_paths, labels))
    if shuffle_files:
      ds = ds.shuffle(len(examples))

    # Fuse load and decode into one function
    def _load_and_decode_fn(*args, **kwargs):
      ex = _load_example(*args, **kwargs)
      return self.info.features.decode_example(ex, decoders=decoders)

    ds = ds.map(
        _load_and_decode_fn, num_parallel_calls=tf.data.experimental.AUTOTUNE
    )
    return ds


def _load_example(
    path: tf.Tensor,
    label: tf.Tensor,
) -> Dict[str, tf.Tensor]:
  img = tf.io.read_file(path)
  return {
      'image': img,
      'label': tf.cast(label, tf.int64),
      'image/filename': path,
  }


def _get_split_label_images(
    root_dir: str,
) -> Tuple[SplitExampleDict, List[str]]:
  """Extract all label names and associated images.

  This function guarantee that examples are deterministically shuffled
  and labels are sorted.

  Args:
    root_dir: The folder where the `split/label/image.png` are located

  Returns:
    split_examples: Mapping split_names -> List[_Example]
    labels: The list off labels
  """
  split_examples = collections.defaultdict(list)
  labels = set()
  for split_name in sorted(_list_folders(root_dir)):
    split_dir = os.path.join(root_dir, split_name)
    for label_name in sorted(_list_folders(split_dir)):
      labels.add(label_name)
      split_examples[split_name].extend(
          [
              _Example(image_path=image_path, label=label_name)
              for image_path in sorted(
                  _list_img_paths(os.path.join(split_dir, label_name))
              )
          ]
      )

  # Shuffle the images deterministically
  for split_name, examples in split_examples.items():
    rgn = random.Random(split_name)  # Uses different seed for each split
    rgn.shuffle(examples)
  return split_examples, sorted(labels)


def _list_folders(root_dir: str) -> List[str]:
  return [
      f
      for f in tf.io.gfile.listdir(root_dir)
      if tf.io.gfile.isdir(os.path.join(root_dir, f))
  ]


def _list_img_paths(root_dir: str) -> List[str]:
  return [
      os.path.join(root_dir, f)
      for f in tf.io.gfile.listdir(root_dir)
      if any(f.lower().endswith(ext) for ext in _SUPPORTED_IMAGE_FORMAT)
  ]



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for tfds.features.Scalar."""

import tensorflow as tf
from tensorflow_datasets import testing
from tensorflow_datasets.core import features


class ScalarFeatureTest(testing.FeatureExpectationsTestCase):

  def test_scalar(self):
    self.assertFeature(
        feature=features.Scalar(dtype=tf.int64, doc='Some description'),
        shape=(),
        dtype=tf.int64,
        tests=[
            testing.FeatureExpectationItem(
                value=42,
                expected=42,
            ),
        ],
    )


if __name__ == '__main__':
  testing.test_main()



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for conllu_dataset_builder_utils."""

from tensorflow_datasets.core.dataset_builders.conll import conllu_dataset_builder_utils as conllu_lib


def test_get_universal_morphology_config():
  lang_1 = "it"
  lang_2 = "de"
  config_1 = conllu_lib.get_universal_morphology_config(
      language=lang_1,
      features=conllu_lib.UNIVERSAL_DEPENDENCIES_FEATURES,
      name=f"test_{lang_1}",
  )
  config_2 = conllu_lib.get_universal_morphology_config(
      language=lang_2,
      features=conllu_lib.UNIVERSAL_DEPENDENCIES_FEATURES,
      name=f"test_{lang_2}",
      description="Config 2 description.",
  )

  assert config_1.name == f"test_{lang_1}"
  assert config_2.name == f"test_{lang_2}"

  assert config_1.language == lang_1
  assert config_2.language == lang_2

  assert config_1.features == config_2.features

  assert config_1.description is None
  assert config_2.description != config_1.description



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for build_community."""
from unittest import mock

from etils import epath

import tensorflow_datasets as tfds  # pylint: disable=unused-import
from tensorflow_datasets.core import naming
from tensorflow_datasets.scripts.documentation import build_community_catalog

DatasetSource = tfds.core.community.dataset_sources.DatasetSource
DatasetPackage = tfds.core.community.register_package.DatasetPackage
DatasetName = naming.DatasetName


def _create_dataset_package(namespace: str, name: str) -> DatasetPackage:
  path = epath.Path(
      f'github://huggingface/datasets/tree/master/datasets/{name}'
  )
  return DatasetPackage(
      name=DatasetName(namespace_name=f'{namespace}:{name}'),
      source=DatasetSource(root_path=path, filenames=[f'{name}.py']),
  )


def _create_templates(
    index_template: str = '',
    dataset_details_template: str = '',
    namespace_toc_template: str = '',
) -> build_community_catalog.DocumentationTemplates:
  return build_community_catalog.DocumentationTemplates(
      index_template=index_template,
      dataset_details_template=dataset_details_template,
      namespace_toc_template=namespace_toc_template,
  )


def _create_options() -> build_community_catalog._Options:
  return build_community_catalog._Options(catalog_dir='/tmp')


def test_dataset_documentation_properties():
  dataset = _create_dataset_package('community1', 'ds1')
  templates = _create_templates()
  dataset_doc = build_community_catalog.DatasetDocumentation(
      dataset=dataset, templates=templates, options=_create_options()
  )
  assert dataset_doc.tfds_id == 'community1:ds1'
  assert dataset_doc.name == 'ds1'


def test_dataset_documentation_links():
  dataset = _create_dataset_package('community1', 'ds1')
  templates = _create_templates()
  dataset_doc = build_community_catalog.DatasetDocumentation(
      dataset=dataset, templates=templates, options=_create_options()
  )
  expected_code_url = (
      '[Code](github://huggingface/datasets/tree/master/datasets/ds1)'
  )
  assert dataset_doc.code_url() == expected_code_url
  assert dataset_doc.extra_links() == [expected_code_url]
  assert (
      dataset_doc.format_extra_links(prefix='* ', infix=' / ')
      == f'* {expected_code_url}'
  )


def test_dataset_documentation_to_details_markdown():
  dataset = _create_dataset_package('community1', 'ds1')
  templates = _create_templates(
      dataset_details_template='{tfds_id}\n{references_bulleted_list}'
  )
  dataset_doc = build_community_catalog.DatasetDocumentation(
      dataset=dataset, templates=templates, options=_create_options()
  )
  assert (
      dataset_doc.to_details_markdown()
      == 'community1:ds1\n*   '
      '[Code](github://huggingface/datasets/tree/master/datasets/ds1)'
  )


def test_dataset_documentation_to_toc_markdown():
  dataset = _create_dataset_package('community1', 'ds1')
  templates = _create_templates()
  dataset_doc = build_community_catalog.DatasetDocumentation(
      dataset=dataset, templates=templates, options=_create_options()
  )
  assert (
      dataset_doc.to_toc_markdown()
      == 'ds1 ([Code](github://huggingface/datasets/tree/master/datasets/ds1))'
  )


def test_huggingface_dataset_documentation():
  dataset = _create_dataset_package('community1', 'ds1')
  templates = _create_templates()
  dataset_doc = build_community_catalog.HuggingfaceDatasetDocumentation(
      dataset=dataset, templates=templates, options=_create_options()
  )
  assert dataset_doc.extra_links() == [
      '[Code](https://github.com/huggingface/datasets/blob/master/datasets/ds1)',
      '[Huggingface](https://huggingface.co/datasets/ds1)',
  ]


def test_huggingface_formatter_overview_page():
  namespace1 = 'community1'
  datasets = []
  templates = _create_templates()
  formatter = build_community_catalog.HuggingfaceFormatter(
      namespace=namespace1,
      datasets=datasets,
      templates=templates,
      options=_create_options(),
  )
  assert formatter.overview_page == 'community1.md'


@mock.patch('tensorflow_datasets.core.github_api.github_path.get_content')
def test_huggingface_formatter_to_namespace_overview(get_content_mock):
  get_content_mock.return_value = '{}'
  namespace1 = 'huggingface'
  datasets = [
      _create_dataset_package(namespace1, 'ds1'),
      _create_dataset_package(namespace1, 'ds2'),
  ]
  templates = _create_templates(dataset_details_template='{tfds_id}')
  get_content_mock.return_value = '{}'
  formatter = build_community_catalog.HuggingfaceFormatter(
      namespace=namespace1,
      datasets=datasets,
      templates=templates,
      options=_create_options(),
  )
  assert (
      formatter.to_namespace_overview()
      == """\
# Huggingface datasets

Huggingface has forked TFDS and provides a lot of text datasets. See
[here](https://huggingface.co/docs/datasets/) for more documentation.
Next you can find the list of all the datasets that can be used with TFDS.


*  [ds1](huggingface/ds1.md)
*  [ds2](huggingface/ds2.md)
"""
  )


@mock.patch('tensorflow_datasets.core.github_api.github_path.get_content')
def test_huggingface_formatter_to_toc_markdown(get_content_mock):
  get_content_mock.return_value = '{}'
  namespace1 = 'community1'
  datasets = [
      _create_dataset_package(namespace1, 'ds1'),
      _create_dataset_package(namespace1, 'ds2'),
  ]
  templates = _create_templates(namespace_toc_template='{name}')

  formatter = build_community_catalog.HuggingfaceFormatter(
      namespace=namespace1,
      datasets=datasets,
      templates=templates,
      options=_create_options(),
  )
  assert formatter.to_toc_markdown() == 'Community1'



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test import."""

import sys
from unittest import mock

from absl import logging
import numpy as np

import unittest


class ImportWithoutTfTest(unittest.TestCase):

  @mock.patch.object(logging, 'log_first_n')
  @mock.patch('builtins.print')
  def test_import_tfds_without_loading_tf(self, print_mock, log_first_n):
    # Check that conftest.py didn't already import TFDS:
    self.assertNotIn('tensorflow_datasets', sys.modules)

    import tensorflow_datasets as tfds  # pylint: disable=g-import-not-at-top

    class DummyDataset(tfds.core.dataset_builder.GeneratorBasedBuilder):
      """Test DatasetBuilder."""

      VERSION = tfds.core.Version('1.0.0')

      def _info(self):
        return tfds.core.dataset_info.DatasetInfo(
            builder=self,
            features=tfds.core.features.FeaturesDict({
                'integer': np.int32,
                'nested': tfds.core.features.FeaturesDict(
                    {'text': tfds.core.features.Text()}
                ),
                'text': tfds.core.features.Text(),
            }),
        )

      def _split_generators(self, dl_manager):
        return {
            'train': self._generate_examples(),
            'test': self._generate_examples(),
        }

      def _generate_examples(self):
        for i in range(20):
          yield i, {
              'integer': i,
              'nested': {'text': 'nested_text'},
              'text': f'test_{i}',
          }

    self.assertNotIn('tensorflow', sys.modules)
    self.assertNotIn('array_record', sys.modules)
    self.assertNotIn('pandas', sys.modules)

    data_dir = '/tmp/import_without_tf'
    builder = DummyDataset(data_dir=data_dir)
    builder.download_and_prepare(
        file_format=tfds.core.file_adapters.FileFormat.ARRAY_RECORD,
    )
    data_source = builder.as_data_source()
    self.assertLen(data_source['train'], 20)
    self.assertEqual(
        data_source['train'][0],
        {
            'integer': 6,
            'nested': {'text': b'nested_text'},
            'text': b'test_6',
        },
    )

    # No warning concerning TensorFlow DTypes was dispatched while loading
    self.assertFalse(log_first_n.called)
    self.assertNotIn('tensorflow', sys.modules)
    self.assertIn('array_record', sys.modules)

    # No error concerning TensorFlow was dispatched:
    self.assertLen(print_mock.call_args_list, 2)
    first_call_args = print_mock.call_args_list[0][0][0]
    self.assertRegex(first_call_args, '.*Downloading and preparing dataset.*')
    second_call_args = print_mock.call_args_list[1][0][0]
    self.assertRegex(second_call_args, '.*Dataset .* downloaded and prepared.*')

    # The function `tfds.testing.mock_data` can be used without TF:
    with tfds.testing.mock_data():
      assert 'tensorflow' not in sys.modules


if __name__ == '__main__':
  unittest.main()



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Wikipedia-based Image Text (WIT) Dataset for the Kaggle competition."""
from __future__ import annotations

import base64
import csv
import functools
import gzip
import io
import sys
from typing import List, Optional

import numpy as np
from tensorflow_datasets.core.utils import bool_utils
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
import tensorflow_datasets.public_api as tfds

_DESCRIPTION = """
Wikipedia - Image/Caption Matching Kaggle Competition.

This competition is organized by the
[Research team](https://research.wikimedia.org/) at the
[Wikimedia Foundation](https://wikimediafoundation.org/) in collaboration with
Google Research and a few external collaborators.
This competition is based on the
[WIT dataset](https://github.com/google-research-datasets/wit) published by
Google Research as detailed in this\
[SIGIR paper](https://dl.acm.org/doi/abs/10.1145/3404835.3463257).

In this competition, you’ll build a model that automatically retrieves the text
closest to an image. Specifically, you'll train your model to associate given
images with article titles or complex captions, in multiple languages.
The best models will account for the semantic granularity of Wikipedia images.
If successful, you'll be contributing to the accessibility of the largest
online encyclopedia. The millions of Wikipedia readers and edietors will be able
to more easily understand, search, and describe media at scale. As a result,
you’ll contribute to an open model to improve learning for all.
"""

_CITATION = """
@article{srinivasan2021wit,
  title={WIT: Wikipedia-based Image Text Dataset for Multimodal Multilingual Machine Learning},
  author={Srinivasan, Krishna and Raman, Karthik and Chen, Jiecao and Bendersky, Michael and Najork, Marc},
  journal={arXiv preprint arXiv:2103.01913},
  year={2021}
}
"""

_EMPTY_IMAGE_BYTES = "iVBORw0KGgoAAAANSUhEUgAAAAEAAAABCAYAAAAfFcSJAAAADUlEQVR42mP8z/C/HgAGgwJ/lK3Q6wAAAABJRU5ErkJggg=="

_BEAM_NAMESPACE = "TFDS_WIT_KAGGLE"


class WitKaggleConfig(tfds.core.BuilderConfig):
  """BuilderConfig for WitKaggle."""

  def __init__(
      self,
      *,
      sample_files: Optional[List[str]] = None,
      image_folder: Optional[str] = None,
      split_specific_features: Optional[tfds.features.FeaturesDict] = None,
      resnet_embedding_shape: int = 2048,
      **kwargs,
  ):
    """BuilderConfig for WitKaggle.

    Args:
      sample_files: list of paths to the sample files.
      image_folder: path to the folder where image data lies.
      split_specific_features: tfds.features.FeaturesDict. The features for the
        specific WitKaggle split.
      resnet_embedding_shape: shape of resnet embeddings.
      **kwargs: keyword arguments forwarded to super.
    """
    super(WitKaggleConfig, self).__init__(**kwargs)
    self.sample_files = sample_files
    self.image_folder = image_folder

    # Features common to all configs.
    common_features = tfds.features.FeaturesDict({
        "caption_title_and_reference_description": (
            tfds.features.Text()
        ),  # Caption for the image_url (if existent, else '').
        "image_url": tfds.features.Text(),
        "image": (
            tfds.features.Image()
        ),  # Base64 encoded image bytes (if existent, else a blank image).
        "metadata_url": (
            tfds.features.Text()
        ),  # Url to the image's commons page (if existent, else '').
        "embedding": tfds.features.Tensor(
            shape=(resnet_embedding_shape,), dtype=np.float32
        ),  # A tensor of 2048 floats (if existent, else zeros).
    })
    self.features = tfds.features.FeaturesDict(
        {**common_features, **split_specific_features}
    )
    self.split_specific_features = split_specific_features
    self.resnet_embedding_shape = resnet_embedding_shape
    # For missing images, we use a zeroed vector of dimensionality (2048,) as
    # resnet embedding.
    self.empty_resnet_embedding = [0] * resnet_embedding_shape


class WitKaggle(tfds.core.GeneratorBasedBuilder):
  """DatasetBuilder for wit_kaggle dataset."""

  VERSION = tfds.core.Version("1.0.2")
  RELEASE_NOTES = {
      "1.0.2": "Fixes parsing of boolean fields.",
      "1.0.1": (
          "Optimize Beam pipeline to avoid strugglers, ignoring rows without "
          "an image URL. Also added more Beam counters."
      ),
      "1.0.0": """Initial release. It provides the train and test datasets from the
      Wikipedia - Image/Caption Matching Kaggle competition
      (https://www.kaggle.com/c/wikipedia-image-caption/data).

      The goal of the competition is to build a model that automatically
      retrieves the text closest to an image. Specifically, the model shuld be
      trained to associate given images with article titles or complex captions,
      in multiple languages. The best models will account for the semantic
      granularity of Wikipedia images.

      Note that this release doesn't provide the ground truth for the test set,
      as it hasn't been provided by the Kaggle competition yet.

      Note that not all of the training observations have corresponding image
      data. The released images exclude all images containing humans. For
      samples which are not associated with image data, the following image
      features are used: `image` is a byte-64 encoded blank image, `embedding`
      is a vector of 2048 zeros.

      The samples released for the competition can be loaded as:
      ```
      tfds.load("wit_kaggle/train_with_extended_features")
      tfds.load("wit_kaggle/test_without_gold")
      ```
      """,
  }

  # pytype: disable=wrong-keyword-args
  # In the Wikipedia - Image/Caption Matching competition, train samples are
  # associated with a rich set of metadata, while test samples only have a
  # sample_id and four image-related fields.
  BUILDER_CONFIGS = [
      WitKaggleConfig(
          name="train_with_extended_features",
          description=(
              "Training samples for the Wikipedia-Image/Caption Matching"
              " competition."
          ),
          sample_files=[
              f"train/train-0000{i}-of-00005.tsv.zip" for i in range(5)
          ],
          image_folder="train/image_data_train",
          split_specific_features=tfds.features.FeaturesDict({
              "language": tfds.features.Text(),
              "page_url": tfds.features.Text(),
              "image_url": tfds.features.Text(),
              "page_title": tfds.features.Text(),
              "section_title": tfds.features.Text(),
              "hierarchical_section_title": tfds.features.Text(),
              "caption_reference_description": tfds.features.Text(),
              "caption_attribution_description": tfds.features.Text(),
              "caption_alt_text_description": tfds.features.Text(),
              "mime_type": tfds.features.Text(),
              "original_height": np.int32,
              "original_width": np.int32,
              "is_main_image": np.bool_,
              "attribution_passes_lang_id": np.bool_,
              "page_changed_recently": np.bool_,
              "context_page_description": tfds.features.Text(),
              "context_section_description": tfds.features.Text(),
              "caption_title_and_reference_description": tfds.features.Text(),
          }),
      ),
      WitKaggleConfig(
          name="test_without_gold",
          description=(
              "Test samples (without gold answers) for the"
              " Wikipedia-Image/Caption Matching competition."
          ),
          sample_files=["test/test.tsv.zip"],
          image_folder="test/image_data_test",
          split_specific_features=tfds.features.FeaturesDict({
              "id": tfds.features.Text(),
              "image_url": tfds.features.Text(),
          }),
      ),
  ]
  # pytype: enable=wrong-keyword-args

  MANUAL_DOWNLOAD_INSTRUCTIONS = """
  Depending on the config called, manual_dir should contain some of the
  following subdirectories:
    * train
      - train-{0000x}-of-00005.tsv.zip
      - image_data_train/
        * image_pixels/
          - train_image_pixels_part-00{000-199}.csv.gz
        * resnet_embeddings/
          - train_resnet_embeddings_part-00{000-214}.csv.gz
    * test
      - test.tsv.zip
      - image_data_test/
        * image_pixels/
          - test_image_pixels_part-0000{0-4}.csv
        * resnet_embeddings/
          - test_resnet_embeddings_part-0000{0-9}.csv

  Registration at https://www.kaggle.com/c/wikipedia-image-caption/data
  is needed to get the links to download the dataset.
  """

  def _info(self) -> tfds.core.DatasetInfo:
    """Returns the dataset metadata."""
    return tfds.core.DatasetInfo(
        builder=self,
        description=_DESCRIPTION,
        features=self.builder_config.features,
        supervised_keys=(
            "image_url",
            "caption_title_and_reference_description",
        ),
        homepage="https://www.kaggle.com/c/wikipedia-image-caption/code",
        citation=_CITATION,
    )

  def _split_generators(self, dl_manager, pipeline):
    """Returns SplitGenerators."""
    archive_path = {
        "samples": [
            tfds.download.Resource(  # pylint:disable=g-complex-comprehension
                path=dl_manager.manual_dir / filename,
                extract_method=tfds.download.ExtractMethod.ZIP,
            )
            for filename in self.builder_config.sample_files
        ],
        "images": tfds.download.Resource(
            path=dl_manager.manual_dir / self.builder_config.image_folder
        ),
    }

    extracted_paths = dl_manager.extract(archive_path)

    return {
        self.builder_config.name: self._generate_examples(
            pipeline=pipeline,
            samples_path=extracted_paths["samples"],
            image_pixels_path=extracted_paths["images"] / "image_pixels",
            image_resnet_path=extracted_paths["images"] / "resnet_embeddings",
        )
    }

  def _generate_examples(
      self, pipeline, samples_path, image_pixels_path, image_resnet_path
  ):
    """Processes the dataset and yields examples.

    Args:
      pipeline: the beam pipeline.
      samples_path: path to the split's sentences.
      image_pixels_path: path to the images' pixel representations.
      image_resnet_path: path to the images' pixel representations.

    Returns:
      Examples.
    """
    beam = tfds.core.lazy_imports.apache_beam
    counter = functools.partial(beam.metrics.Metrics.counter, _BEAM_NAMESPACE)

    def _get_csv_reader(filename):
      if filename.suffix == ".gz":
        counter("gz_csv_files").inc()
        g = tf.io.gfile.GFile(filename, "rb")
        f = gzip.open(g, "rt", newline="")
      else:
        counter("normal_csv_files").inc()
        f = tf.io.gfile.GFile(filename, "r")
      # Limit to 100 MB. Value must be smaller than the C long maximum value.
      csv.field_size_limit(sys.maxsize)
      return csv.reader(f, delimiter="\t")

    def _read_pixel_rows(filename):
      r"""Contains image_url \t image_pixel \t metadata_url."""
      reader = _get_csv_reader(filename)
      for row in reader:
        counter("pixel_rows").inc()
        image_url, image_representation, metadata_url = row
        if image_url:
          yield [image_url, (image_representation, metadata_url)]
        else:
          counter("pixel_rows_no_image_url").inc()

    def _read_resnet_rows(filename):
      r"""Contains image_url \t resnet_embedding."""
      reader = _get_csv_reader(filename)
      for row in reader:
        counter("resnet_rows").inc()
        image_url, image_representation = row
        if image_url:
          yield [image_url, image_representation]
        else:
          counter("resnet_rows_no_image_url").inc()

    def _read_samples_rows(folder_path):
      """Contains samples: train and test have different fields."""
      for filename in tf.io.gfile.listdir(folder_path):
        file_path = folder_path / filename
        f = tf.io.gfile.GFile(file_path, "r")
        # Limit to 100 MB. Value must be smaller than the C long maximum value.
        csv.field_size_limit(sys.maxsize)
        csv_reader = csv.DictReader(f, delimiter="\t", quoting=csv.QUOTE_ALL)
        for row in csv_reader:
          counter("samples_rows").inc()
          sample = {
              feature_key: row[feature_key]
              for feature_key in self.builder_config.split_specific_features.keys()
          }
          image_url = row["image_url"]
          if image_url:
            yield [image_url, sample]
          else:
            counter("samples_rows_no_image_url").inc()

    def _process_examples(el):
      sample_url, sample_fields = el
      # Each image_url can be associated with multiple samples (e.g., multiple
      # languages).
      for i, sample_info in enumerate(sample_fields["sample_info"]):
        sample_id = f"{i}_{sample_url}"
        sample = {"image_url": sample_url}
        for feature_key in self.builder_config.split_specific_features.keys():
          sample[feature_key] = sample_info[feature_key]
          is_boolean_feature = (
              self.builder_config.split_specific_features[feature_key].np_dtype
              == np.bool_
          )
          if is_boolean_feature:
            sample[feature_key] = bool_utils.parse_bool(sample[feature_key])
        # Test samples don't have gold captions.
        if "caption_title_and_reference_description" not in sample_info:
          sample["caption_title_and_reference_description"] = ""

        # We output image data only if there is at least one image
        # representation per image_url.
        # Not all of the samples in the competition have corresponding image
        # data. In case multiple different image representations are associated
        # with the same image_url, we don't know which one is correct and don't
        # output any.
        if len(set(sample_fields["image_pixels"])) == 1:
          sample_image, sample_metadata = sample_fields["image_pixels"][0]
          sample["image"] = io.BytesIO(base64.b64decode(sample_image))
          sample["metadata_url"] = sample_metadata
        else:
          if len(set(sample_fields["image_pixels"])) > 1:
            counter("image_pixels_multiple").inc()
          else:
            counter("image_pixels_missing").inc()
            sample["image"] = io.BytesIO(base64.b64decode(_EMPTY_IMAGE_BYTES))
          sample["metadata_url"] = ""

        if len(set(sample_fields["image_resnet"])) == 1:
          image_resnet = [
              float(x) for x in sample_fields["image_resnet"][0].split(",")
          ]
          sample["embedding"] = image_resnet
        else:
          if len(set(sample_fields["image_resnet"])) > 1:
            counter("image_resnet_multiple").inc()
          else:
            counter("image_resnet_missing").inc()
          sample["embedding"] = self.builder_config.empty_resnet_embedding

        yield sample_id, sample

    # Read embeddings and bytes representations from (possibly compressed) csv.
    image_resnet_files = [
        image_resnet_path / f for f in tf.io.gfile.listdir(image_resnet_path)
    ]
    resnet_collection = (
        pipeline
        | "Collection from resnet files" >> beam.Create(image_resnet_files)
        | "Get embeddings per image" >> beam.FlatMap(_read_resnet_rows)
    )

    image_pixel_files = [
        image_pixels_path / f for f in tf.io.gfile.listdir(image_pixels_path)
    ]
    pixel_collection = (
        pipeline
        | "Collection from pixel files" >> beam.Create(image_pixel_files)
        | "Get pixels per image" >> beam.FlatMap(_read_pixel_rows)
    )

    # Read samples from tsv files.
    sample_collection = (
        pipeline
        | "Collection from sample files" >> beam.Create(samples_path)
        | "Get samples" >> beam.FlatMap(_read_samples_rows)
    )

    # Combine the features and yield examples.
    return (
        {
            "sample_info": sample_collection,
            "image_pixels": pixel_collection,
            "image_resnet": resnet_collection,
        }
        | "Group by image_url" >> beam.CoGroupByKey()
        | "Reshuffle" >> beam.Reshuffle()
        | "Process and yield examples" >> beam.FlatMap(_process_examples)
    )



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Partial feature connector decoding util."""

import typing
from typing import Any, Callable, Optional, Union

from tensorflow_datasets.core import features as features_lib
from tensorflow_datasets.core import utils
from tensorflow_datasets.core.decode import base
from tensorflow_datasets.core.features import features_dict
from tensorflow_datasets.core.utils import dtype_utils

# Expected feature specs provided by the user
_FeatureSpecElem = Union[features_lib.FeatureConnector, Any]
FeatureSpecs = utils.TreeDict[_FeatureSpecElem]

DecoderArg = Union[base.Decoder, 'PartialDecoding']


class PartialDecoding:
  """Allow to decode a subset of features.

  See guide:
  https://www.tensorflow.org/datasets/decode#only_decode_a_sub-set_of_the_features
  """

  def __init__(
      self,
      features: FeatureSpecs,
      decoders: Optional[utils.TreeDict[base.Decoder]] = None,
  ):
    """Constructor.

    Args:
      features: A nested dict of `tfds.features.FeatureConnector`
      decoders: Optional additional decoders to apply (e.g.
        `tfds.decode.SkipDecoding()`)
    """
    self.decoders = decoders
    self._feature_specs = features  # Expected feature specs

  def extract_features(
      self,
      features: features_lib.FeatureConnector,
  ) -> features_lib.FeatureConnector:
    """Returns the `tfds.features.FeaturesDict`.

    Extract the subset of features

    Args:
      features: Features on which extract the sub-set

    Returns:
      features_subset: A subset of the features
    """
    with utils.try_reraise(
        'Provided PartialDecoding specs does not match actual features: '
    ):
      # Convert non-features into features
      expected_feature = _normalize_feature_item(
          feature=features,
          expected_feature=self._feature_specs,
      )
      # Get the intersection of `features` and `expected_feature`
      return _extract_features(
          feature=features,
          expected_feature=features_dict.to_feature(expected_feature),
      )


def _normalize_feature_item(
    feature: features_lib.FeatureConnector,
    expected_feature: FeatureSpecs,
) -> FeatureSpecs:
  """Extracts the features matching the expected_feature structure."""
  # If user provide a FeatureConnector, use this
  if (isinstance(expected_feature, features_lib.FeatureConnector)) or (
      dtype_utils.is_np_or_tf_dtype(expected_feature)
  ):
    return expected_feature
  # If the user provide a bool, use the matching feature connector
  # Example: {'cameras': True} -> `{'camera': FeatureDict({'image': Image()})}`
  elif isinstance(expected_feature, bool):
    assert expected_feature  # `False` values should have been filtered already
    return feature
  # If the user provide a sequence, merge it with the associated feature.
  elif isinstance(expected_feature, (list, set, dict)):
    if isinstance(expected_feature, (list, set)):
      expected_feature = {k: True for k in expected_feature}
    return _normalize_feature_dict(
        feature=feature,
        expected_feature=expected_feature,
    )
  else:
    raise TypeError(f'Unexpected partial feature spec: {expected_feature!r}')


def _normalize_feature_dict(
    feature: features_lib.FeatureConnector,
    expected_feature: FeatureSpecs,
) -> FeatureSpecs:
  """Extract the features matching the expected_feature structure."""
  if type(feature) == features_lib.FeaturesDict:  # pylint: disable=unidiomatic-typecheck
    inner_features = {
        k: v for k, v in expected_feature.items() if v is not False  # pylint: disable=g-bool-id-comparison
    }
    inner_features = {  # Extract the feature subset  # pylint: disable=g-complex-comprehension
        k: _extract_feature_item(  # pytype: disable=wrong-arg-types  # always-use-return-annotations
            feature=feature,
            expected_key=k,
            expected_value=v,
            fn=_normalize_feature_item,
        )
        for k, v in inner_features.items()
    }
    # Filter `False` values
    return inner_features
  elif type(feature) == features_lib.Sequence:  # pylint: disable=unidiomatic-typecheck
    inner_features = _normalize_feature_dict(
        feature=feature.feature,  # pytype: disable=attribute-error
        expected_feature=expected_feature,
    )
    return features_lib.Sequence(inner_features)
  else:
    raise ValueError(
        f'Unexpected structure {expected_feature!r} does not match {feature!r}'
    )


def _extract_features(
    feature: features_lib.FeatureConnector,
    expected_feature: features_lib.FeatureConnector,
) -> features_lib.FeatureConnector:
  """Recursive implementation of `PartialDecoding.extract_features`."""
  # Feature types should match
  if not isinstance(feature, type(expected_feature)):
    raise TypeError(f'Expected: {expected_feature}. Got: {feature}')

  # Recurse into FeaturesDict, Sequence
  # Use `type` rather than `isinstance` to not recurse into inherited classes.
  if type(feature) == features_lib.FeaturesDict:  # pylint: disable=unidiomatic-typecheck
    expected_feature = typing.cast(features_lib.FeaturesDict, expected_feature)
    return features_lib.FeaturesDict(
        {  # Extract the feature subset  # pylint: disable=g-complex-comprehension
            k: _extract_feature_item(  # pytype: disable=wrong-arg-types  # always-use-return-annotations
                feature=feature,
                expected_key=k,
                expected_value=v,
                fn=_extract_features,
            )
            for k, v in expected_feature.items()
        }
    )
  elif type(feature) == features_lib.Sequence:  # pylint: disable=unidiomatic-typecheck
    feature = typing.cast(features_lib.Sequence, feature)
    expected_feature = typing.cast(features_lib.Sequence, expected_feature)
    feature_subset = _extract_features(
        feature=feature.feature,
        expected_feature=expected_feature.feature,
    )
    return features_lib.Sequence(feature_subset, length=feature._length)  # pylint: disable=protected-access
  else:
    # Assert that the specs matches
    if (
        feature.dtype != expected_feature.dtype
        or not utils.shapes_are_compatible(
            feature.shape, expected_feature.shape
        )
    ):
      raise ValueError(f'Expected: {expected_feature}. Got: {feature}')
    return feature


def _extract_feature_item(
    feature: features_lib.FeaturesDict,
    expected_key: str,
    expected_value: features_lib.FeatureConnector,
    fn: Callable[..., Any],
) -> features_lib.FeatureConnector:
  """Calls `_extract_features(feature[key], expected_feature=value)`."""
  assert isinstance(feature, features_lib.FeaturesDict)
  if expected_key not in feature:
    raise ValueError(f'Missing expected feature {expected_key!r}.')

  with utils.try_reraise(f'In {expected_key!r}: '):
    return fn(feature=feature[expected_key], expected_feature=expected_value)



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Dmlab dataset."""

from __future__ import annotations

import io
import os

from absl import logging
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
import tensorflow_datasets.public_api as tfds

_URL = "https://storage.googleapis.com/dmlab-vtab/dmlab.tar.gz"


class Dmlab(tfds.core.GeneratorBasedBuilder):
  """Dmlab dataset."""

  VERSION = tfds.core.Version("2.0.1")

  def _info(self):
    return tfds.core.DatasetInfo(
        builder=self,
        description=r"""
        The Dmlab dataset contains frames observed by the agent acting in the
        DeepMind Lab environment, which are annotated by the distance between
        the agent and various objects present in the environment. The goal is to
        is to evaluate the ability of a visual model to reason about distances
        from the visual input in 3D environments. The Dmlab dataset consists of
        360x480 color images in 6 classes. The classes are
        {close, far, very far} x {positive reward, negative reward}
        respectively.""",
        features=tfds.features.FeaturesDict({
            "image": tfds.features.Image(
                shape=(360, 480, 3), encoding_format="jpeg"
            ),
            "filename": tfds.features.Text(),
            "label": tfds.features.ClassLabel(num_classes=6),
        }),
        homepage="https://github.com/google-research/task_adaptation",
        citation=r"""@article{zhai2019visual,
        title={The Visual Task Adaptation Benchmark},
        author={Xiaohua Zhai and Joan Puigcerver and Alexander Kolesnikov and
               Pierre Ruyssen and Carlos Riquelme and Mario Lucic and
               Josip Djolonga and Andre Susano Pinto and Maxim Neumann and
               Alexey Dosovitskiy and Lucas Beyer and Olivier Bachem and
               Michael Tschannen and Marcin Michalski and Olivier Bousquet and
               Sylvain Gelly and Neil Houlsby},
                              year={2019},
                              eprint={1910.04867},
                              archivePrefix={arXiv},
                              primaryClass={cs.CV},
                              url = {https://arxiv.org/abs/1910.04867}
                          }""",
        supervised_keys=("image", "label"),
    )

  def _split_generators(self, dl_manager):
    path = dl_manager.download_and_extract(_URL)

    return [
        tfds.core.SplitGenerator(
            name=tfds.Split.TRAIN,
            gen_kwargs={
                "images_dir_path": path,
                "split_name": "train",
            },
        ),
        tfds.core.SplitGenerator(
            name=tfds.Split.VALIDATION,
            gen_kwargs={
                "images_dir_path": path,
                "split_name": "validation",
            },
        ),
        tfds.core.SplitGenerator(
            name=tfds.Split.TEST,
            gen_kwargs={
                "images_dir_path": path,
                "split_name": "test",
            },
        ),
    ]

  def _parse_single_image(self, example_proto):
    """Parses single video from the input tfrecords.

    Args:
      example_proto: tfExample proto with a single video.

    Returns:
      dict with all frames, positions and actions.
    """

    feature_map = {
        "image": tf.io.FixedLenFeature(shape=[], dtype=tf.string),
        "filename": tf.io.FixedLenFeature(shape=[], dtype=tf.string),
        "label": tf.io.FixedLenFeature(shape=[], dtype=tf.int64),
    }

    parse_single = tf.io.parse_single_example(example_proto, feature_map)

    return parse_single

  def _generate_examples(self, images_dir_path, split_name):
    path_glob = os.path.join(
        images_dir_path, "dmlab-{}.tfrecord*".format(split_name)
    )
    files = tf.io.gfile.glob(path_glob)

    logging.info("Reading data from %s.", ",".join(files))
    with tf.Graph().as_default():
      ds = tf.data.TFRecordDataset(files)
      ds = ds.map(
          self._parse_single_image,
          num_parallel_calls=tf.data.experimental.AUTOTUNE,
      )
      iterator = tf.compat.v1.data.make_one_shot_iterator(ds).get_next()
      with tf.compat.v1.Session() as sess:
        sess.run(tf.compat.v1.global_variables_initializer())
        try:
          while True:
            result = sess.run(iterator)
            yield result["filename"], {
                "image": io.BytesIO(result["image"]),
                "filename": result["filename"],
                "label": result["label"],
            }

        except tf.errors.OutOfRangeError:
          return



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Templates to generate dataset builders.

Depending on the given format, it will generate code to add format-specific
dataset builders to tfds.
"""
import textwrap

from tensorflow_datasets.scripts.cli import cli_utils

# Names of dataset builders.
STANDARD = 'standard'
CONLL = 'conll'
CONLLU = 'conllu'


def create_builder_template(info: cli_utils.DatasetInfo):
  """Calls the required (possibly format-specific) dataset builder template.

  Args:
    info: `cli_utils.DatasetInfo` object containing all information about the
      dataset necessary to generate a template for the new dataset.

  Returns:
    A string containing the builder-specific template for the new dataset, to be
    filled by the user.
  """
  if info.data_format == STANDARD:
    return _standard_template(info)
  elif info.data_format == CONLL:
    return _conll_template(info)
  elif info.data_format == CONLLU:
    return _conllu_template(info)
  else:
    raise ValueError(
        f"Required format {info.data_format} isn't associated with "
        'a format-specific builder in TFDS.'
    )


def _standard_template(info: cli_utils.DatasetInfo) -> str:
  """Returns a template for a `tfds.core.GeneratorBasedBuilder`."""
  content = textwrap.dedent(
      f'''\
      """{info.name} dataset."""

      import {info.tfds_api} as tfds


      class Builder(tfds.core.GeneratorBasedBuilder):
        """DatasetBuilder for {info.name} dataset."""

        VERSION = tfds.core.Version('1.0.0')
        RELEASE_NOTES = {{
            '1.0.0': 'Initial release.',
        }}

        def _info(self) -> tfds.core.DatasetInfo:
          """Returns the dataset metadata."""
          # {info.todo}: Specifies the tfds.core.DatasetInfo object
          return self.dataset_info_from_configs(
              features=tfds.features.FeaturesDict({{
                  # These are the features of your dataset like images, labels ...
                  'image': tfds.features.Image(shape=(None, None, 3)),
                  'label': tfds.features.ClassLabel(names=['no', 'yes']),
              }}),
              # If there's a common (input, target) tuple from the
              # features, specify them here. They'll be used if
              # `as_supervised=True` in `builder.as_dataset`.
              supervised_keys=('image', 'label'),  # Set to `None` to disable
              homepage='https://dataset-homepage/',
          )

        def _split_generators(self, dl_manager: tfds.download.DownloadManager):
          """Returns SplitGenerators."""
          # {info.todo}: Downloads the data and defines the splits
          path = dl_manager.download_and_extract('https://todo-data-url')

          # {info.todo}: Returns the Dict[split names, Iterator[Key, Example]]
          return {{
              'train': self._generate_examples(path / 'train_imgs'),
          }}

        def _generate_examples(self, path):
          """Yields examples."""
          # {info.todo}: Yields (key, example) tuples from the dataset
          for f in path.glob('*.jpeg'):
            yield 'key', {{
                'image': f,
                'label': 'yes',
            }}
      '''
  )
  return content


def _conll_template(info: cli_utils.DatasetInfo) -> str:
  """A template for ConllDatasetBuilder."""

  content = textwrap.dedent(
      f'''\
      """{info.name} dataset."""

      from tensorflow_datasets.core.dataset_builders.conll import conll_dataset_builder_utils as conll_lib
      import {info.tfds_api} as tfds


      class {info.cls_name}(tfds.dataset_builders.ConllDatasetBuilder):
        """DatasetBuilder for {info.name} dataset."""

        VERSION = tfds.core.Version('1.0.0')
        RELEASE_NOTES = {{
            '1.0.0': 'Initial release.',
        }}
        # {info.todo}: Add details about the dataset's features.
        # conll_lib contains a set of ready-to-use features.
        BUILDER_CONFIGS = [conll_lib.CONLL_2003_CONFIG]

        def _info(self) -> tfds.core.DatasetInfo:
          """Returns the dataset metadata."""
          # {info.todo}: Specifies the dataset infos.
          return self.create_dataset_info(
            homepage="",
          )

        def _split_generators(self, dl_manager: tfds.download.DownloadManager):
          """Returns SplitGenerators."""
          # {info.todo}: Downloads the data and defines the splits
          path = dl_manager.download_and_extract('https://todo-data-url')

          return {{
            'train': self._generate_examples(path / 'train.txt',)
          }}

        # {info.todo}: If you need a customized _generate_examples function,
        # comment out the following.
        # def _generate_examples(
        #     self,
        #     path: Union[epath.PathLike, List[epath.PathLike]],
        #   ):
        #   """Yields (key, example) examples."""
        #   pass
      '''
  )
  return content


def _conllu_template(info: cli_utils.DatasetInfo) -> str:
  """A template for ConllUDatasetBuilder."""

  content = textwrap.dedent(
      f'''\
      """{info.name} dataset."""

      from tensorflow_datasets.core.dataset_builders.conll import conllu_dataset_builder_utils as conllu_lib
      import {info.tfds_api} as tfds


      class {info.cls_name}(tfds.dataset_builders.ConllUDatasetBuilder):
        """DatasetBuilder for {info.name} dataset."""

        VERSION = tfds.core.Version('1.0.0')
        RELEASE_NOTES = {{
            '1.0.0': 'Initial release.',
        }}
        # {info.todo}: Add details about the dataset's config to use to
        # BUILDER_CONFIGS. conllu_lib contains a set of ready-to-use features.
        BUILDER_CONFIGS = [
            conllu_lib.get_universal_morphology_config(
                language='',
                features=conllu_lib.UNIVERSAL_DEPENDENCIES_FEATURES,
                name='')
        ]

        def _info(self) -> tfds.core.DatasetInfo:
          """Returns the dataset metadata."""
          # {info.todo}: Specifies the dataset infos.
          return self.create_dataset_info(
              homepage='',
          )

        def _split_generators(self, dl_manager: tfds.download.DownloadManager):
          """Returns SplitGenerators."""
          # {info.todo}: Downloads the data and defines the splits
          path = dl_manager.download_and_extract('https://todo-data-url')

          # {info.todo}: Specify the process_example_fn to be used in the
          # `_generate_examples` method to process examples.
          # The default `process_example_fn` processes a conllu-annotated
          # example using the features specified in BUILDER_CONFIGS.
          # `conllu_dataset_builder` already provides a number of ready-to-use
          # process functions.
          return {{
              'train':
                  self._generate_examples(
                      filepaths=path / 'train.txt',
                      # Remove if you want to use the default `process_example_fn`.
                      # process_example_fn=
                  )
          }}

        # {info.todo}: If you need a customized _generate_examples function,
        # comment out the following, otherwise remove it.
        # def _generate_examples(
        #     self,
        #     path: Union[epath.PathLike, List[epath.PathLike]],
        #   ):
        #   """Yields (key, example) examples."""
        #   pass
      '''
  )
  return content



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for tensorflow_datasets.core.dataset_utils."""

from etils import enp
import numpy as np
import tensorflow as tf
from tensorflow_datasets import testing
from tensorflow_datasets.core import dataset_utils


def _create_dataset(rng):
  return tf.data.Dataset.from_tensor_slices(list(rng))


class DatasetAsNumPyTest(testing.TestCase):

  @testing.run_in_graph_and_eager_modes()
  def test_singleton_tensor(self):
    t = tf.random.normal((10, 10))
    np_t = dataset_utils.as_numpy(t)
    self.assertEqual((10, 10), np_t.shape)
    self.assertEqual(np.float32, np_t.dtype)

  @testing.run_in_graph_and_eager_modes()
  def test_nested_tensors(self):
    t1 = tf.random.normal((10, 10))
    t2 = tf.random.normal((10, 20))
    nest_tup = (t1, t2)
    np_t1, np_t2 = dataset_utils.as_numpy(nest_tup)
    self.assertEqual((10, 10), np_t1.shape)
    self.assertEqual(np.float32, np_t1.dtype)
    self.assertEqual((10, 20), np_t2.shape)
    self.assertEqual(np.float32, np_t2.dtype)

    nest_dict = {"foo": t1, "bar": {"zoo": t2}}
    np_nest_dict = dataset_utils.as_numpy(nest_dict)
    np_t1 = np_nest_dict["foo"]
    np_t2 = np_nest_dict["bar"]["zoo"]
    self.assertEqual((10, 10), np_t1.shape)
    self.assertEqual(np.float32, np_t1.dtype)
    self.assertEqual((10, 20), np_t2.shape)
    self.assertEqual(np.float32, np_t2.dtype)

  @testing.run_in_graph_and_eager_modes()
  def test_singleton_dataset(self):
    ds = _create_dataset(range(10))
    np_ds = dataset_utils.as_numpy(ds)
    self.assertEqual(list(range(10)), [int(el) for el in list(np_ds)])
    # Iterating twice on the dataset recreate the iterator.
    self.assertEqual(list(range(10)), [int(el) for el in list(np_ds)])

    if tf.executing_eagerly():
      self.assertEqual(len(np_ds), 10)
      self.assertEqual(
          np_ds.element_spec,
          enp.ArraySpec(shape=(), dtype=np.int32),
      )
    else:
      with self.assertRaisesWithPredicateMatch(
          TypeError, "__len__() is not supported for `tfds.as_numpy`"
      ):
        _ = len(np_ds)

  def test_with_graph(self):
    with tf.Graph().as_default():
      ds = _create_dataset(range(10))
      np_ds = dataset_utils.as_numpy(ds)
    self.assertEqual(list(range(10)), [int(el) for el in list(np_ds)])

  @testing.run_in_graph_and_eager_modes()
  def test_singleton_dataset_with_nested_elements(self):
    ds = _create_dataset(range(10))
    ds = ds.map(lambda el: {"a": el, "b": el + 1, "c": (el + 2, el + 3)})
    np_ds = dataset_utils.as_numpy(ds)
    for i, el in enumerate(np_ds):
      self.assertEqual(i, el["a"])
      self.assertEqual(i + 1, el["b"])
      self.assertEqual(i + 2, el["c"][0])
      self.assertEqual(i + 3, el["c"][1])

    if tf.executing_eagerly():
      self.assertEqual(
          np_ds.element_spec,
          {
              "a": enp.ArraySpec(shape=(), dtype=np.int32),
              "b": enp.ArraySpec(shape=(), dtype=np.int32),
              "c": (
                  enp.ArraySpec(shape=(), dtype=np.int32),
                  enp.ArraySpec(shape=(), dtype=np.int32),
              ),
          },
      )

  @testing.run_in_graph_and_eager_modes()
  def test_nested_dataset_sequential_access(self):
    ds1 = _create_dataset(range(10))
    ds2 = _create_dataset(range(10, 20))
    np_ds = dataset_utils.as_numpy((ds1, {"a": ds2}))
    np_ds1 = np_ds[0]
    np_ds2 = np_ds[1]["a"]

    self.assertEqual(list(range(10)), [int(el) for el in list(np_ds1)])
    self.assertEqual(list(range(10, 20)), [int(el) for el in list(np_ds2)])

  @testing.run_in_graph_and_eager_modes()
  def test_nested_dataset_simultaneous_access(self):
    ds1 = _create_dataset(range(10))
    ds2 = _create_dataset(range(10, 20))
    np_ds = dataset_utils.as_numpy((ds1, {"a": ds2}))
    np_ds1 = np_ds[0]
    np_ds2 = np_ds[1]["a"]

    for i1, i2 in zip(np_ds1, np_ds2):
      self.assertEqual(i2, int(i1) + 10)

  @testing.run_in_graph_and_eager_modes()
  def test_nested_dataset_nested_elements(self):
    ds1 = _create_dataset(range(10))
    ds1 = ds1.map(lambda el: {"a": el, "b": el + 1, "c": (el + 2, el + 3)})
    ds2 = _create_dataset(range(10, 20))
    np_ds = dataset_utils.as_numpy((ds1, {"a": ds2}))
    np_ds1 = np_ds[0]
    np_ds2 = np_ds[1]["a"]

    for i, (el1, el2) in enumerate(zip(np_ds1, np_ds2)):
      self.assertEqual(i + 10, el2)
      self.assertEqual(i, el1["a"])
      self.assertEqual(i + 1, el1["b"])
      self.assertEqual(i + 2, el1["c"][0])
      self.assertEqual(i + 3, el1["c"][1])

  @testing.run_in_graph_and_eager_modes()
  def test_tensors_match(self):
    t = tf.random.uniform(
        shape=(50, 3),
        maxval=1000,
        dtype=tf.int32,
    )

    ds = dataset_utils.as_numpy({"a": t, "b": t})
    # sess.run() should be called a single time for all input. Otherwise input
    # and target may not match
    self.assertAllEqual(ds["a"], ds["b"])

  @testing.run_in_graph_and_eager_modes()
  def test_ragged_tensors(self):
    rt = tf.ragged.constant([
        [1, 2, 3],
        [],
        [4, 5],
    ])
    rt = dataset_utils.as_numpy(rt)

    if not tf.executing_eagerly():
      # Output of `sess.run(rt)` is a `RaggedTensorValue` object
      self.assertIsInstance(rt, tf.compat.v1.ragged.RaggedTensorValue)
    else:
      self.assertIsInstance(rt, tf.RaggedTensor)

    self.assertAllEqual(
        rt,
        tf.ragged.constant([
            [1, 2, 3],
            [],
            [4, 5],
        ]),
    )

  @testing.run_in_graph_and_eager_modes()
  def test_ragged_tensors_ds(self):
    def _gen_ragged_tensors():
      # Yield the (flat_values, rowids)
      yield ([0, 1, 2, 3], [0, 0, 0, 2])  # ex0
      yield ([], [])  # ex1
      yield ([4, 5, 6], [0, 1, 1])  # ex2

    ds = tf.data.Dataset.from_generator(
        _gen_ragged_tensors,
        output_types=(tf.int64, tf.int64),
        output_shapes=((None,), (None,)),
    )
    ds = ds.map(tf.RaggedTensor.from_value_rowids)

    rt0, rt1, rt2 = list(dataset_utils.as_numpy(ds))
    self.assertAllEqual(
        rt0,
        [
            [0, 1, 2],
            [],
            [
                3,
            ],
        ],
    )
    self.assertAllEqual(rt1, [])
    self.assertAllEqual(rt2, [[4], [5, 6]])

  def test_none_ds(self):
    ds = tf.data.Dataset.range(10)
    ds = ds.map(lambda x: (x, None))
    exs = list(dataset_utils.as_numpy(ds))
    self.assertAllEqual(exs, [(x, None) for x in range(10)])


if __name__ == "__main__":
  testing.test_main()



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""yahoo_ltrc dataset."""

import dataclasses

import numpy as np
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
import tensorflow_datasets.public_api as tfds
from tensorflow_datasets.ranking.libsvm_ranking_parser import LibSVMRankingParser

_DESCRIPTION = """
The Yahoo Learning to Rank Challenge dataset (also called "C14") is a
Learning-to-Rank dataset released by Yahoo. The dataset consists of
query-document pairs represented as feature vectors and corresponding relevance
judgment labels.

The dataset contains two versions:

 * `set1`: Containing 709,877 query-document pairs.
 * `set2`: Containing 172,870 query-document pairs.

You can specify whether to use the `set1` or `set2` version of the dataset as
follows:

```python
ds = tfds.load("yahoo_ltrc/set1")
ds = tfds.load("yahoo_ltrc/set2")
```

If only `yahoo_ltrc` is specified, the `yahoo_ltrc/set1` option is selected by
default:

```python
# This is the same as `tfds.load("yahoo_ltrc/set1")`
ds = tfds.load("yahoo_ltrc")
```
"""

_CITATION = """
@inproceedings{chapelle2011yahoo,
  title={Yahoo! learning to rank challenge overview},
  author={Chapelle, Olivier and Chang, Yi},
  booktitle={Proceedings of the learning to rank challenge},
  pages={1--24},
  year={2011},
  organization={PMLR}
}
"""

_LABEL_NAME = "label"


@dataclasses.dataclass
class YahooLTRCConfig(tfds.core.BuilderConfig):
  prefix: str = "set1"
  num_features: int = 699


class YahooLTRC(tfds.core.GeneratorBasedBuilder):
  """DatasetBuilder for yahoo_ltrc dataset."""

  VERSION = tfds.core.Version("1.1.0")
  RELEASE_NOTES = {
      "1.0.0": "Initial release.",
      "1.1.0": "Add query and document identifiers.",
  }
  # pytype: disable=wrong-keyword-args
  # pylint: disable=unexpected-keyword-arg
  BUILDER_CONFIGS = [
      YahooLTRCConfig(name="set1", prefix="set1", num_features=699),
      YahooLTRCConfig(name="set2", prefix="set2", num_features=700),
  ]

  # pylint: enable=unexpected-keyword-arg
  # pytype: enable=wrong-keyword-args

  MANUAL_DOWNLOAD_INSTRUCTIONS = """
  Request access for the C14 Yahoo Learning To Rank Challenge dataset on
  https://research.yahoo.com/datasets. Extract the downloaded `dataset.tgz` file
  and place the `ltrc_yahoo.tar.bz2` file in `manual_dir/`.
  """

  def _info(self) -> tfds.core.DatasetInfo:
    """Returns the dataset metadata."""
    encoding = tfds.features.Encoding.ZLIB
    features = {
        "float_features": tfds.features.Tensor(
            shape=(None, self.builder_config.num_features),
            dtype=np.float64,
            encoding=encoding,
        ),
        _LABEL_NAME: tfds.features.Tensor(
            shape=(None,), dtype=np.float64, encoding=encoding
        ),
        "query_id": tfds.features.Text(),
        "doc_id": tfds.features.Tensor(
            shape=(None,), dtype=np.int64, encoding=encoding
        ),
    }

    return tfds.core.DatasetInfo(
        builder=self,
        description=_DESCRIPTION,
        features=tfds.features.FeaturesDict(features),
        homepage="https://research.yahoo.com/datasets",
        citation=_CITATION,
    )

  def _split_generators(self, dl_manager: tfds.download.DownloadManager):
    """Returns SplitGenerators."""
    # Extract the manually downloaded `ltrc_yahoo.tar.bz2`
    archive_path = dl_manager.manual_dir / "ltrc_yahoo.tar.bz2"
    path = dl_manager.extract(archive_path)

    prefix = self.builder_config.prefix
    feature_names = {
        n: f"feature_{n}"
        for n in range(1, self.builder_config.num_features + 1)
    }

    splits = {
        "train": self._generate_examples(
            path / f"{prefix}.train.txt", feature_names
        ),
        "vali": self._generate_examples(
            path / f"{prefix}.valid.txt", feature_names
        ),
        "test": self._generate_examples(
            path / f"{prefix}.test.txt", feature_names
        ),
    }

    return splits

  def _generate_examples(self, path, feature_names):
    """Yields examples."""
    with tf.io.gfile.GFile(path, "r") as f:
      yield from LibSVMRankingParser(
          f, feature_names, _LABEL_NAME, combine_features=True
      )



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Writer to sequentially write examples to disk."""

from __future__ import annotations

import dataclasses
import os
from typing import Any, Dict, List, Optional

from tensorflow_datasets.core import dataset_info
from tensorflow_datasets.core import example_serializer
from tensorflow_datasets.core import features as features_lib
from tensorflow_datasets.core import file_adapters
from tensorflow_datasets.core import naming
from tensorflow_datasets.core import splits as splits_lib
from tensorflow_datasets.core.utils import py_utils
from tensorflow_datasets.core.utils.lazy_imports_utils import array_record_module
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf



def _serialize_example(
    example: Any,
    features: features_lib.FeaturesDict,
    serializer: example_serializer.ExampleSerializer,
) -> str:
  try:
    encoded_example = features.encode_example(example)
  except Exception as e:  # pylint: disable=broad-except
    py_utils.reraise(
        e, prefix='Failed to encode example:\n', suffix=f'{example}\n'
    )

  return serializer.serialize_example(encoded_example)


@dataclasses.dataclass
class Shard(object):
  """Shard that is being written."""

  writer: tf.io.TFRecordWriter  # TODO(sabela): use a file adapter
  num_examples: int = 0
  num_bytes: int = 0

  def add_example(self, serialized_example: str) -> None:
    """Writes a new example."""
    self.writer.write(serialized_example)
    self.num_examples += 1
    self.num_bytes += len(serialized_example)

  def close_writer(self) -> None:
    """Closes the writer."""
    if hasattr(self.writer, 'flush'):
      self.writer.flush()
    self.writer.close()


@dataclasses.dataclass
class Split(object):
  """Information of a split that is being created."""

  info: splits_lib.SplitInfo
  current_shard: Optional[Shard] = None
  complete_shards: int = 0
  # The dataset name is taken from the builder class.
  ds_name: str = ''
  closed: bool = False
  file_format: file_adapters.FileFormat = file_adapters.FileFormat.TFRECORD

  def add_example(self, serialized_example: str) -> None:
    """Adds an example to the shard.

    If there is no open shard, it starts a new one.

    Args:
      serialized_example: example to add to the shard.

    Raises:
      ValueError: if the split is already closed.
    """
    if self.closed:
      raise ValueError(f'Split {self.info.name} is already closed.')
    if self.current_shard is None:
      if self.info.filename_template is None:
        raise ValueError(f'Split {self.info.name} has no filename template.')
      path = self.info.filename_template.sharded_filepath(
          shard_index=self.complete_shards, num_shards=None
      )
      if self.file_format == file_adapters.FileFormat.TFRECORD:
        self.current_shard = Shard(writer=tf.io.TFRecordWriter(os.fspath(path)))
      elif self.file_format == file_adapters.FileFormat.ARRAY_RECORD:
        self.current_shard = Shard(
            writer=array_record_module.ArrayRecordWriter(
                os.fspath(path), 'group_size:1'
            )
        )
      else:
        raise ValueError('Unknown file format %s' % self.file_format)
    self.current_shard.add_example(serialized_example)

  def close_shard(self) -> None:
    """Finalizes a shard and updates the split metadata accordingly."""
    if not self.current_shard:
      return
    self.current_shard.close_writer()
    self.info = splits_lib.SplitInfo(
        name=self.info.name,
        shard_lengths=self.info.shard_lengths
        + [self.current_shard.num_examples],
        num_bytes=self.info.num_bytes + self.current_shard.num_bytes,
        filename_template=self.info.filename_template,
    )
    self.complete_shards += 1
    self.current_shard = None

  def close(self) -> None:
    """Closes the split (and the shard if it is still open)."""
    if not self.closed and self.current_shard:
      self.close_shard()
    self.closed = True


def _split_dict(splits: Dict[str, Split]) -> splits_lib.SplitDict:
  return splits_lib.SplitDict([split.info for _, split in splits.items()])


def _initialize_split(
    split_name: str,
    data_directory: Any,
    ds_name: str,
    filetype_suffix: str,
    file_format: file_adapters.FileFormat,
    shard_lengths: Optional[List[int]] = None,
    num_bytes: int = 0,
) -> Split:
  """Initializes a split.

  Args:
    split_name: name of the split.
    data_directory: directory where the split data will be located.
    ds_name: name of the dataset.
    filetype_suffix: file format.
    file_format: An entry in file_adapters.FileFormat.
    shard_lengths: if the split already has shards, it contains the list of the
      shard lenghts. If None, it assumes that the split is empty.
    num_bytes: number of bytes that have been written already.

  Returns:
    A Split.
  """
  if not shard_lengths:
    shard_lengths = []
  filename_template = naming.ShardedFileTemplate(
      dataset_name=ds_name,
      data_dir=data_directory,
      split=split_name,
      filetype_suffix=filetype_suffix,
      template='{DATASET}-{SPLIT}.{FILEFORMAT}-{SHARD_INDEX}',
  )
  return Split(
      info=splits_lib.SplitInfo(
          name=split_name,
          shard_lengths=shard_lengths,
          num_bytes=num_bytes,
          filename_template=filename_template,
      ),
      complete_shards=len(shard_lengths),
      ds_name=ds_name,
      file_format=file_format,
  )


class SequentialWriter:
  """Class to write a TFDS dataset sequentially.

  The SequentialWriter can be used to generate TFDS datasets by directly
  appending TF Examples to the desired splits.

  Once the user creates a SequentialWriter with a given DatasetInfo, they can
  create splits, append examples to them, and close them whenever they are
  finished.

  Note that:
  * Not closing a split may cause data to be lost.
  * The examples are written to disk in the same order that they are given to
    the writer.
  * Since the SequentialWriter doesn't know how many examples are going to be
    written, it can't estimate the optimal number of shards per split. Use the
    `max_examples_per_shard` parameter in the constructor to control how many
    elements there should be per shard.

  The datasets written with this writer can be read directly with
  `tfds.builder_from_directories`.

  Example:

  writer = SequentialWriter(ds_info=ds_info, max_examples_per_shard=1000)
  writer.initialize_splits(['train', 'test'])

  while (...):
    # Code that generates the examples
    writer.add_examples({'train': [example1, example2],
                         'test':  [example3]})
    ...

  writer.close_splits()
  """

  # TODO(sabela): add support for beam.
  # TODO(sabela): add support for build_configs.
  # TODO(sabela): support non-TFRecord writers. At the moment, the FileAdapters
  # API only suports writing a set of examples, so the support for other formats
  # would have to be manual.
  def __init__(
      self,
      ds_info: dataset_info.DatasetInfo,
      max_examples_per_shard: int,
      overwrite: bool = True,
      file_format: str = 'tfrecord',
  ):
    """Creates a SequentialWriter.

    Args:
      ds_info: DatasetInfo for this dataset.
      max_examples_per_shard: maximum number of examples to write per shard.
      overwrite: if True, it ignores and overwrites any existing data.
        Otherwise, it loads the existing dataset and appends the new data (new
        data will always be created as new shards).
      file_format: An entry in file_adapters.FileFormat.
    """

    self._data_dir = ds_info.data_dir
    self._ds_name = ds_info.name
    self._ds_info = ds_info
    self._file_format = file_adapters.FileFormat.from_value(file_format)
    if not overwrite:
      try:
        self._ds_info.read_from_directory(self._data_dir)
        # read_from_directory does some checks but not on the dataset name.
        if self._ds_info.name != self._ds_name:
          raise ValueError(
              f'Trying to append a dataset with name {ds_info.name}'
              f' to an existing dataset with name {self._ds_info.name}'
          )
      except FileNotFoundError:
        self._ds_info.set_file_format(
            file_format=self._file_format,
            # if it was set, we want this to fail to warn the user
            override=False,
        )
    else:
      self._ds_info.set_file_format(
          file_format=self._file_format,
          # if it was set, we want this to fail to warn the user
          override=False,
      )

    self._filetype_suffix = (
        ds_info.file_format or file_adapters.FileFormat.TFRECORD
    ).file_suffix
    self._max_examples_per_shard = max_examples_per_shard
    self._splits = {}
    if not overwrite:
      for split_name, split in ds_info.splits.items():
        self._splits[split_name] = _initialize_split(
            split_name=split_name,
            data_directory=self._data_dir,
            ds_name=self._ds_name,
            filetype_suffix=self._filetype_suffix,
            shard_lengths=split.shard_lengths,
            num_bytes=split.num_bytes,
            file_format=self._file_format,
        )
    self._serializer = example_serializer.ExampleSerializer(
        self._ds_info.features.get_serialized_info()
    )

  def initialize_splits(
      self, splits: List[str], fail_if_exists: bool = True
  ) -> None:
    """Adds new splits to the dataset.

    Args:
      splits: list of split names to add.
      fail_if_exists: will fail if this split already contains data.

    Raises:
      KeyError: if the split is already present.
    """
    for split in splits:
      if split in self._splits:
        if fail_if_exists:
          raise KeyError(f'Split {split} was already initialized.')
        else:
          continue
      self._splits[split] = _initialize_split(
          split_name=split,
          data_directory=self._data_dir,
          ds_name=self._ds_name,
          filetype_suffix=self._filetype_suffix,
          file_format=self._file_format,
      )
    self._write_splits_metadata()

  def add_examples(self, split_examples: Dict[str, List[Any]]) -> None:
    """Adds examples to the splits.

    Args:
      split_examples: dictionary of `split_name`:list_of_examples that includes
        the list of examples that has to be added to each of the splits. Not all
        the existing splits have to be in the dictionary

    Raises:
      KeyError: if any of the splits doesn't exist.
    """
    # Write the example and update the shard
    for split_name, examples in split_examples.items():
      split = self._splits.get(split_name, None)
      if not split:
        raise KeyError(f'Split {split} was not initialized.')
      for example in examples:
        serialized_example = _serialize_example(
            example, self._ds_info.features, self._serializer
        )
        split.add_example(serialized_example)
        if split.current_shard.num_examples >= self._max_examples_per_shard:
          split.close_shard()
    self._write_splits_metadata()

  def _write_splits_metadata(self) -> None:
    self._ds_info.set_splits(_split_dict(self._splits))
    # Either writes the first metadata, or overwrites it.
    self._ds_info.write_to_directory(self._data_dir)

  def close_splits(self, splits: List[str]) -> None:
    """Closes the given list of splits.

    Args:
      splits: list of split names.

    Raises:
      KeyError: if any of the splits doesn't exist.
    """
    for split_name in splits:
      if split_name not in self._splits:
        raise KeyError(f'Split {split_name} was not created.')
      split = self._splits[split_name]
      split.close()
    self._write_splits_metadata()

  def close_all(self) -> None:
    """Closes all the open splits."""
    for _, split in self._splits.items():
      split.close()
    self._write_splits_metadata()



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Cats vs Dogs dataset."""

import re

from absl import logging
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
import tensorflow_datasets.public_api as tfds

_CITATION = """\
@Inproceedings (Conference){asirra-a-captcha-that-exploits-interest-aligned-manual-image-categorization,
author = {Elson, Jeremy and Douceur, John (JD) and Howell, Jon and Saul, Jared},
title = {Asirra: A CAPTCHA that Exploits Interest-Aligned Manual Image Categorization},
booktitle = {Proceedings of 14th ACM Conference on Computer and Communications Security (CCS)},
year = {2007},
month = {October},
publisher = {Association for Computing Machinery, Inc.},
url = {https://www.microsoft.com/en-us/research/publication/asirra-a-captcha-that-exploits-interest-aligned-manual-image-categorization/},
edition = {Proceedings of 14th ACM Conference on Computer and Communications Security (CCS)},
}
"""

_URL = (
    "https://download.microsoft.com/download/3/E/1/"
    "3E1C3F21-ECDB-4869-8368-6DEBA77B919F/kagglecatsanddogs_5340.zip"
)
_NUM_CORRUPT_IMAGES = 1738
_DESCRIPTION = (
    "A large set of images of cats and dogs. "
    "There are %d corrupted images that are dropped." % _NUM_CORRUPT_IMAGES
)

_NAME_RE = re.compile(r"^PetImages[\\/](Cat|Dog)[\\/]\d+\.jpg$")


class CatsVsDogs(tfds.core.GeneratorBasedBuilder):
  """Cats vs Dogs."""

  VERSION = tfds.core.Version("4.0.0")
  RELEASE_NOTES = {
      "4.0.0": "New split API (https://tensorflow.org/datasets/splits)",
  }

  def _info(self):
    return tfds.core.DatasetInfo(
        builder=self,
        description=_DESCRIPTION,
        features=tfds.features.FeaturesDict({
            "image": tfds.features.Image(),
            "image/filename": tfds.features.Text(),  # eg 'PetImages/Dog/0.jpg'
            "label": tfds.features.ClassLabel(names=["cat", "dog"]),
        }),
        supervised_keys=("image", "label"),
        homepage=(
            "https://www.microsoft.com/en-us/download/details.aspx?id=54765"
        ),
        citation=_CITATION,
    )

  def _split_generators(self, dl_manager):
    path = dl_manager.download(_URL)

    # There is no predefined train/val/test split for this dataset.
    return [
        tfds.core.SplitGenerator(
            name=tfds.Split.TRAIN,
            gen_kwargs={
                "archive": dl_manager.iter_archive(path),
            },
        ),
    ]

  def _generate_examples(self, archive):
    """Generate Cats vs Dogs images and labels given a directory path."""
    num_skipped = 0
    for fname, fobj in archive:
      res = _NAME_RE.match(fname)
      if not res:  # README file, ...
        continue
      label = res.group(1).lower()
      if tf.compat.as_bytes("JFIF") not in fobj.peek(10):
        num_skipped += 1
        continue
      record = {
          "image": fobj,
          "image/filename": fname,
          "label": label,
      }
      yield fname, record

    if num_skipped != _NUM_CORRUPT_IMAGES:
      raise ValueError(
          "Expected %d corrupt images, but found %d"
          % (_NUM_CORRUPT_IMAGES, num_skipped)
      )
    logging.warning("%d images were corrupted and were skipped", num_skipped)



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""GTZAN Music Speech dataset."""

import os
import re

from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
import tensorflow_datasets.public_api as tfds

_CITATION = """
@ONLINE {Music Speech,
    author = "Tzanetakis, George",
    title  = "GTZAN Music/Speech Collection",
    year   = "1999",
    url    = "http://marsyas.info/index.html"
}
"""

_DESCRIPTION = """
The dataset was collected for the purposes of music/speech discrimination.
The dataset consists of 120 tracks, each 30 seconds long.
Each class (music/speech) has 60 examples.
The tracks are all 22050Hz Mono 16-bit audio files in .wav format.
"""

_DOWNLOAD_URL = "http://opihi.cs.uvic.ca/sound/music_speech.tar.gz"
_HOMEPAGE_URL = "http://marsyas.info/index.html"

_CLASS_LABELS = ["music", "speech"]


class GTZANMusicSpeech(tfds.core.GeneratorBasedBuilder):
  """GTZANMusicSpeech Dataset."""

  VERSION = tfds.core.Version("1.0.0")

  def _info(self):
    return tfds.core.DatasetInfo(
        builder=self,
        description=_DESCRIPTION,
        features=tfds.features.FeaturesDict({
            "audio": tfds.features.Audio(file_format="wav", sample_rate=22050),
            "label": tfds.features.ClassLabel(names=_CLASS_LABELS),
            "audio/filename": tfds.features.Text(),
        }),
        supervised_keys=("audio", "label"),
        homepage=_HOMEPAGE_URL,
        citation=_CITATION,
    )

  def _split_generators(self, dl_manager):
    """Returns SplitGenerators."""
    dl_paths = dl_manager.download_and_extract({"music_speech": _DOWNLOAD_URL})
    path = os.path.join(dl_paths["music_speech"], "music_speech")
    # There is no predefined train/val/test split for this dataset.
    return [
        tfds.core.SplitGenerator(
            name=tfds.Split.TRAIN, gen_kwargs={"path": path}
        ),
    ]

  def _generate_examples(self, path):
    """Yields examples.

    Args:
       path: Path of the downloaded and extracted directory

    Yields:
       Next examples
    """

    #  The wav files are in directories named "{label}_wav"
    name_regex = re.compile(r"(.*)(speech|music)\_.*\/[\w-]*\.(wav)")

    for root, _, file_name in tf.io.gfile.walk(path):
      for fname in file_name:
        full_file_name = os.path.join(root, fname)
        res = name_regex.match(full_file_name)
        if not res:
          continue
        label = res.group(2)
        key = fname
        example = {
            "audio": full_file_name,
            "label": label,
            "audio/filename": fname,
        }
        yield key, example



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""covid19 dataset."""

import tensorflow_datasets.public_api as tfds
from tensorflow_datasets.structured.covid19 import covid19


class Covid19Test(tfds.testing.DatasetBuilderTestCase):
  """Tests for covid-19 dataset."""

  DATASET_CLASS = covid19.Covid19
  SPLITS = {
      'train': 3,
  }
  DL_EXTRACT_RESULT = 'covid19.csv.gz'

  def setUp(self):
    super().setUp()
    covid19._N_RECORDS = 3
    covid19._BATCH_SIZE = 2


if __name__ == '__main__':
  tfds.testing.test_main()



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for tensorflow_datasets.core.community.dataset_sources."""

import os

from tensorflow_datasets.core import github_api
from tensorflow_datasets.core.community import dataset_sources


def test_dataset_source():
  uri = 'github://owner/repo/tree/master/.../audio/gtzan/gtzan.py'
  src = dataset_sources.DatasetSource.from_json(uri)
  assert isinstance(src.root_path, github_api.GithubPath)
  assert (
      os.fspath(src.root_path)
      == 'github://owner/repo/tree/master/.../audio/gtzan'
  )
  assert src.filenames == ['gtzan.py']
  src_json = src.to_json()
  assert uri == src_json
  assert src == dataset_sources.DatasetSource.from_json(src_json)


def test_dataset_source_multifiles():
  json_input = {
      'root_path': 'github://owner/repo/tree/master/.../audio/gtzan',
      'filenames': ['checksums.tsv', 'gtzan.py'],
  }
  src = dataset_sources.DatasetSource.from_json(json_input)
  assert isinstance(src.root_path, github_api.GithubPath)
  assert (
      os.fspath(src.root_path)
      == 'github://owner/repo/tree/master/.../audio/gtzan'
  )
  assert src.filenames == ['checksums.tsv', 'gtzan.py']
  src_json = src.to_json()
  assert json_input == src_json
  assert src == dataset_sources.DatasetSource.from_json(src_json)



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utils to handle resources."""

from etils import epath

to_write_path = epath.to_write_path


def tfds_path(*relative_path: epath.PathLike) -> epath.Path:
  """Path to `tensorflow_datasets/` root dir.

  The following examples are equivalent:

  ```py
  path = tfds.core.tfds_path() / 'path/to/data.txt'
  path = tfds.core.tfds_path('path/to/data.txt')
  path = tfds.core.tfds_path('path', 'to', 'data.txt')
  ```

  Note: Even if `/` is used, those examples are compatible with Windows, as
  pathlib will automatically normalize the paths.

  Args:
    *relative_path: Relative path, eventually to concatenate.

  Returns:
    path: The root TFDS path.
  """
  return epath.resource_path('tensorflow_datasets').joinpath(*relative_path)


def tfds_write_path(*relative_path: epath.PathLike) -> epath.Path:
  """Path to `tensorflow_datasets/` root dir (read-write).

  Contrary to `tfds.core.tfds_path`, path returned here support write
  operations. Used in scripts to update the TFDS repository.

  Args:
    *relative_path: Relative path, eventually to concatenate.

  Returns:
    path: The root TFDS path.
  """
  return epath.to_write_path(tfds_path(*relative_path))



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for tensorflow_datasets.core.load.

Note: `load.py` code was previously in `registered.py`, so some of the tests
are still on `registered_test.py`.
"""

from unittest import mock

import pytest
import tensorflow as tf
from tensorflow_datasets import testing
from tensorflow_datasets.core import file_adapters
from tensorflow_datasets.core import load
from tensorflow_datasets.core import naming
from tensorflow_datasets.core import read_only_builder
from tensorflow_datasets.core import registered
from tensorflow_datasets.core import visibility
def test_load_hf_dataset():
  builder = object()
  with mock.patch(
      'tensorflow_datasets.core.dataset_builders.huggingface_dataset_builder.builder',
      return_value=builder,
  ):
    assert load.builder('huggingface:x/y') is builder


@visibility.set_availables_tmp([
    visibility.DatasetType.COMMUNITY_PUBLIC,
])
def test_community_public_load():
  with mock.patch(
      'tensorflow_datasets.core.community.community_register.list_builders',
      return_value=['ns:ds'],
  ), mock.patch(
      'tensorflow_datasets.core.community.community_register.builder_cls',
      return_value=testing.DummyDataset,
  ), mock.patch(
      'tensorflow_datasets.core.registered.list_imported_builders',
      return_value=[],
  ):
    assert load.list_builders() == ['ns:ds']

    # Builder is correctly returned
    assert load.builder_cls('ns:ds') is testing.DummyDataset
    assert isinstance(load.builder('ns:ds'), testing.DummyDataset)


@pytest.fixture(scope='session')
def dummy_dc_loader() -> load.DatasetCollectionLoader:
  return load.DatasetCollectionLoader(
      collection=testing.DummyDatasetCollection()
  )


def test_dc_loader_name(dummy_dc_loader: load.DatasetCollectionLoader):  # pylint: disable=redefined-outer-name
  assert dummy_dc_loader.collection_name == 'dummy_dataset_collection'


def test_load_dataset(dummy_dc_loader: load.DatasetCollectionLoader):  # pylint: disable=redefined-outer-name
  with mock.patch.object(load, 'load', autospec=True) as mock_load:
    examples = tf.data.Dataset.from_tensor_slices([1, 2, 3])
    expected = {'train': examples, 'test': examples}
    mock_load.return_value = expected
    loaded_dataset = dummy_dc_loader.load_dataset('c')
    mock_load.assert_called_once_with(name='c/e:3.5.7', with_info=False)
    assert loaded_dataset == expected


def test_load_dataset_split(dummy_dc_loader: load.DatasetCollectionLoader):  # pylint: disable=redefined-outer-name
  with mock.patch.object(load, 'load', autospec=True) as mock_load:
    examples = tf.data.Dataset.from_tensor_slices([1, 2, 3])
    expected = {'train': examples}
    mock_load.return_value = [examples, examples]
    loaded_dataset = dummy_dc_loader.load_dataset('c', split='train')
    mock_load.assert_called_once_with(
        name='c/e:3.5.7', with_info=False, split=['train']
    )
    assert loaded_dataset == expected


def test_load_dataset_splits(dummy_dc_loader: load.DatasetCollectionLoader):  # pylint: disable=redefined-outer-name
  with mock.patch.object(load, 'load', autospec=True) as mock_load:
    examples = tf.data.Dataset.from_tensor_slices([1, 2, 3])
    expected = {'train': examples, 'test': examples}
    mock_load.return_value = [examples, examples]
    loaded_dataset = dummy_dc_loader.load_dataset('c', split=['train', 'test'])
    mock_load.assert_called_once_with(
        name='c/e:3.5.7', with_info=False, split=['train', 'test']
    )
    assert loaded_dataset == expected


def test_load_dataset_runtime_error(
    dummy_dc_loader: load.DatasetCollectionLoader,
):  # pylint: disable=redefined-outer-name
  with pytest.raises(RuntimeError, match='Unsupported return type.+'):
    with mock.patch.object(load, 'load', autospec=True) as mock_load:
      examples = tf.data.Dataset.from_tensor_slices([1, 2, 3])
      mock_load.return_value = examples
      dummy_dc_loader.load_dataset('c')


def test_load_dataset_key_error(dummy_dc_loader: load.DatasetCollectionLoader):  # pylint: disable=redefined-outer-name
  with pytest.raises(
      KeyError, match='Dataset d is not included in this collection.+'
  ):
    dummy_dc_loader.load_dataset('d')


def test_load_dataset_with_kwargs(
    dummy_dc_loader: load.DatasetCollectionLoader,
):  # pylint: disable=redefined-outer-name
  with mock.patch.object(load, 'load', autospec=True) as mock_load:
    examples = tf.data.Dataset.from_tensor_slices([1, 2, 3])
    expected = {'train': examples, 'test': examples}
    mock_load.return_value = expected
    loaded_dataset = dummy_dc_loader.load_dataset(
        'c', loader_kwargs={'with_info': True, 'batch_size': 3}
    )

    mock_load.assert_called_once_with(
        name='c/e:3.5.7', with_info=False, batch_size=3
    )
    assert loaded_dataset == expected


@pytest.mark.parametrize(
    'builder_kwargs',
    [
        None,
        {'file_format': 'array_record'},
        {'file_format': file_adapters.FileFormat.ARRAY_RECORD},
    ],
)
def test_data_source_defaults_to_array_record_format(
    builder_kwargs,
):
  with mock.patch.object(load, 'builder', autospec=True) as mock_builder:
    load.data_source(
        'mydataset', builder_kwargs=builder_kwargs,
    )
    mock_builder.assert_called_with(
        'mydataset',
        data_dir=None,
        try_gcs=False,
        file_format=file_adapters.FileFormat.ARRAY_RECORD,
    )
@pytest.mark.parametrize(
    'file_format',
    ['tfrecord', file_adapters.FileFormat.TFRECORD],
)
def test_data_source_raises_error_for_other_file_formats(file_format):
  with pytest.raises(
      NotImplementedError, match='No random access data source'
  ):
    with mock.patch.object(load, 'builder', autospec=True):
      load.data_source(
          'mydataset', builder_kwargs={'file_format': file_format}
      )



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""As dataframe util."""

from __future__ import annotations

import dataclasses
import typing
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import numpy as np
from tensorflow_datasets.core import dataset_info
from tensorflow_datasets.core import dataset_utils
from tensorflow_datasets.core import features
from tensorflow_datasets.core import lazy_imports_lib
from tensorflow_datasets.core.utils import py_utils
from tensorflow_datasets.core.utils import type_utils
from tensorflow_datasets.core.utils.lazy_imports_utils import pandas as pd
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf

# Should be `pandas.io.formats.style.Styler`, but is a costly import
Styler = Any

TreeDict = type_utils.TreeDict


@dataclasses.dataclass
class ColumnInfo:
  """`pandas.DataFrame` columns info (name, style formating,...).

  Attributes:
    name: Name of the column
    format_fn: Function applied to each column items, which returns the
      displayed string object (eventually HTML)
  """

  name: str
  format_fn: Optional[Callable[[np.ndarray], str]] = None
  # Should also add a `style.apply` function

  @classmethod
  def from_spec(
      cls,
      path: Tuple[str],
      ds_info: Optional[dataset_info.DatasetInfo],
  ) -> 'ColumnInfo':
    """Formatter which filters values hard to read and format."""
    name = '/'.join(path)

    # If ds_info is not provided, no formatting
    if not ds_info:
      # Could use the spec for a better formatting ?
      return cls(name)

    # Extract feature for formatting
    feature, sequence_rank = _get_feature(path, ds_info.features)

    # Sequence would require special formatting, not supported for now
    if sequence_rank == 0:
      repr_fn = feature.repr_html
    elif sequence_rank == 1:
      repr_fn = feature.repr_html_batch
    elif sequence_rank > 1:
      repr_fn = feature.repr_html_ragged

    def repr_fn_with_debug(val):  # Wrap repr_fn to add debug info
      try:
        return repr_fn(val)
      except Exception as e:  # pylint: disable=broad-except
        err_msg = (
            f'HTML formatting of column {name} failed:\n'
            f' * feature: {feature}\n'
            f' * input: {val!r}\n'
        )
        py_utils.reraise(e, prefix=err_msg)

    return ColumnInfo(
        name='/'.join(path),
        format_fn=repr_fn_with_debug,
    )


def _get_feature(
    path: Tuple[str, ...],
    feature: features.FeatureConnector,
) -> Tuple[features.FeatureConnector, int]:
  """Recursively extracts the feature and sequence rank (plain, ragged, ...)."""
  sequence_rank = 0

  # Collapse the nested sequences
  while isinstance(feature, features.Sequence):
    # Subclasses like `Video` shouldn't be recursed into.
    # But sequence of dict like `TranslationVariableLanguages` should.
    # Currently, there is no good way for a composed sub-feature to only
    # display a single column instead of one per sub-feature.
    # So `MyFeature({'x': tf.int32, 'y': tf.bool})` will have 2 columns `x`
    # and `y`.
    if type(feature) != features.Sequence and not path:  # pylint: disable=unidiomatic-typecheck
      break
    sequence_rank += 1
    feature = feature.feature  # Extract inner feature  # pytype: disable=attribute-error

  if path:  # Has level deeper, recurse
    feature = typing.cast(features.FeaturesDict, feature)
    feature, nested_sequence_rank = _get_feature(path[1:], feature[path[0]])  # pytype: disable=wrong-arg-types
    sequence_rank += nested_sequence_rank

  return feature, sequence_rank


def _make_columns(
    specs: TreeDict[tf.TypeSpec],
    ds_info: Optional[dataset_info.DatasetInfo],
) -> List[ColumnInfo]:
  """Extract the columns info of the `panda.DataFrame`."""
  return [
      ColumnInfo.from_spec(path, ds_info)
      for path, _ in py_utils.flatten_with_path(specs)
  ]


def _make_row_dict(
    ex: TreeDict[np.ndarray],
    columns: List[ColumnInfo],
) -> Dict[str, np.ndarray]:
  """Convert a single example into a `pandas.DataFrame` row."""
  values = tf.nest.flatten(ex)
  return {column.name: v for column, v in zip(columns, values)}


def as_dataframe(
    ds: tf.data.Dataset,
    ds_info: Optional[dataset_info.DatasetInfo] = None,
) -> pd.DataFrame:
  """Convert the dataset into a pandas dataframe.

  Warning: The dataframe will be loaded entirely in memory, you may
  want to call `tfds.as_dataframe` on a subset of the data instead:

  ```
  df = tfds.as_dataframe(ds.take(10), ds_info)
  ```

  Args:
    ds: `tf.data.Dataset`. The tf.data.Dataset object to convert to panda
      dataframe. Examples should not be batched. The full dataset will be
      loaded.
    ds_info: Dataset info object. If given, helps improving the formatting.
      Available either through `tfds.load('mnist', with_info=True)` or
      `tfds.builder('mnist').info`

  Returns:
    dataframe: The `pandas.DataFrame` object
  """
  # Raise a clean error message if panda isn't installed.
  lazy_imports_lib.lazy_imports.pandas  # pylint: disable=pointless-statement

  class StyledDataFrame(pd.DataFrame):
    """`pandas.DataFrame` displayed as `pandas.io.formats.style.Styler`.

    `StyledDataFrame` is a `pandas.DataFrame` with better Jupyter notebook
    representation. Contrary to regular `pandas.DataFrame`, the `style` is
    attached to the `pandas.DataFrame`.

    ```
    df = StyledDataFrame(...)
    df.current_style.apply(...)  # Configure the style
    df  # The data-frame is displayed using ` pandas.io.formats.style.Styler`
    ```
    """

    # StyledDataFrame could be improved such as the style is forwarded when
    # selecting sub-data frames.

    def __init__(self, *args, **kwargs):
      super().__init__(*args, **kwargs)  # pytype: disable=wrong-arg-count  # re-none
      # Use name-mangling for forward-compatibility in case pandas
      # adds a `_styler` attribute in the future.
      self.__styler: Optional[Styler] = None

    @property
    def current_style(self) -> Styler:
      """Like `pandas.DataFrame.style`, but attach the style to the DataFrame."""
      if self.__styler is None:
        self.__styler = super().style  # pytype: disable=attribute-error  # re-none
      return self.__styler

    def _repr_html_(self) -> Union[None, str]:
      # See base class for doc
      if self.__styler is None:
        return super()._repr_html_()  # pytype: disable=attribute-error  # re-none
      return self.__styler._repr_html_()  # pylint: disable=protected-access

  # Pack `as_supervised=True` datasets
  if ds_info:
    ds = dataset_info.pack_as_supervised_ds(ds, ds_info)

  # Flatten the keys names, specs,... while keeping the feature key definition
  # order
  columns = _make_columns(ds.element_spec, ds_info=ds_info)
  rows = [_make_row_dict(ex, columns) for ex in dataset_utils.as_numpy(ds)]
  df = StyledDataFrame(rows)
  df.current_style.format({c.name: c.format_fn for c in columns if c.format_fn})
  return df



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for dataset_metadata."""

from tensorflow_datasets import testing
from tensorflow_datasets.core import dataset_metadata
from tensorflow_datasets.core.utils import resource_utils


class DatasetMetadataTest(testing.TestCase):

  def test_load(self):
    datadir_path = resource_utils.tfds_path(
        "testing/dummy_config_based_datasets/dummy_ds_1"
    )
    metadata = dataset_metadata.load(datadir_path)
    self.assertEqual(
        metadata.description,
        "Description of `dummy_ds_1` dummy config-based dataset.\n",
    )
    self.assertEqual(
        metadata.citation,
        """@Article{google22tfds,
author = "The TFDS team",
title = "TFDS: a collection of ready-to-use datasets for use with TensorFlow, Jax, and other Machine Learning frameworks.",
journal = "ML gazette",
year = "2022"
}""",
    )
    self.assertEqual(metadata.tags, ["content.data-type.image"])

  def test_valid_tags(self):
    valid_tags = dataset_metadata.valid_tags()
    self.assertIn("content.data-type.image", valid_tags)

  def test_valid_tags_with_comments(self):
    text = dataset_metadata.valid_tags_with_comments()
    self.assertIn("content.data-type.image # Contains image data.", text)


if __name__ == "__main__":
  testing.main()



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Smoke Test for docs generation."""

import os
import shutil
import tempfile

from absl.testing import absltest

from tensorflow_datasets.core.utils import docs
from tensorflow_datasets.scripts.documentation import build_api_docs


class BuildDocsTest(absltest.TestCase):

  def setUp(self):
    super(BuildDocsTest, self).setUp()
    self.workdir = tempfile.mkdtemp()
    if os.path.exists(self.workdir):
      shutil.rmtree(self.workdir)
    os.makedirs(self.workdir)

  def test_api_gen(self):
    build_api_docs.execute(
        output_dir=self.workdir,
        code_url_prefix="",
        search_hints=True,
        site_path="datasets/api_docs/python",
    )

    # Check that the "defined in" section is working
    with open(os.path.join(self.workdir, "tfds.md")) as f:
      content = f.read()
    self.assertIn("__init__.py", content)

    # Check that the `testing` folder is generated.
    with open(os.path.join(self.workdir, "tfds/testing.md")) as f:
      content = f.read()
    self.assertIn("__init__.py", content)

  def test_tensorflow_decorators_are_used_and_not_dummy_decorator(self):
    self.assertEqual(docs.deprecated.__name__, "set_deprecated")
    self.assertEqual(docs.doc_private.__name__, "doc_private")
    self.assertEqual(docs.do_not_doc.__name__, "do_not_generate_docs")
    self.assertEqual(
        docs.do_not_doc_inheritable.__name__, "do_not_doc_inheritable"
    )


if __name__ == "__main__":
  absltest.main()



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections

from tensorflow_datasets import testing
from tensorflow_datasets.core import dataset_utils
from tensorflow_datasets.video import ucf101


class Ucf101Test(testing.DatasetBuilderTestCase):
  DATASET_CLASS = ucf101.Ucf101

  SPLITS = {
      'train': 3,
      'test': 2,
  }

  DL_EXTRACT_RESULT = {
      'videos': 'videos',
      'splits': 'splits',
  }

  BUILDER_CONFIG_NAMES_TO_TEST = ['ucf101_1_256', 'ucf101_2']

  def _assertAsDataset(self, builder):
    """Check the label distribution for each split."""
    super(Ucf101Test, self)._assertAsDataset(builder)
    label_frequncies = {}
    label_feature = builder.info.features['label']
    dataset = builder.as_dataset()
    for split_name in Ucf101Test.SPLITS:
      label_frequncies[split_name] = collections.defaultdict(int)
      for features in dataset_utils.as_numpy(dataset[split_name]):
        label_name = label_feature.int2str(features['label'])
        label_frequncies[split_name][label_name] += 1
    self.assertEqual(
        dict(label_frequncies),
        {
            'test': {'Archery': 1, 'Nunchucks': 1},
            'train': {'Archery': 1, 'Nunchucks': 2},
        },
    )


if __name__ == '__main__':
  testing.test_main()



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Czech Restaurants dataset."""

import tensorflow_datasets.public_api as tfds
from tensorflow_datasets.structured.cs_restaurants import cs_restaurants


class CSRestaurantsTest(tfds.testing.DatasetBuilderTestCase):
  DATASET_CLASS = cs_restaurants.CSRestaurants
  SPLITS = {
      'train': 3,
      'validation': 2,
      'test': 2,
  }

  DL_EXTRACT_RESULT = {
      'train_path': 'train.json',
      'dev_path': 'dev.json',
      'test_path': 'test.json',
  }

  def test_get_table_from_da(self):
    da = 'inform(food=Mexican,kids_allowed=no,name=Ferdinanda)'
    self.assertCountEqual(
        cs_restaurants._get_table_from_da(da),
        [
            {
                'column_header': 'intent',
                'row_number': 1,
                'content': 'inform',
            },
            {
                'column_header': 'food',
                'row_number': 1,
                'content': 'Mexican',
            },
            {
                'column_header': 'kids_allowed',
                'row_number': 1,
                'content': 'no',
            },
            {
                'column_header': 'name',
                'row_number': 1,
                'content': 'Ferdinanda',
            },
        ],
    )


if __name__ == '__main__':
  tfds.testing.test_main()



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Wiki40B: A clean Wikipedia dataset for 40+ languages."""

import os

from absl import logging
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
import tensorflow_datasets.public_api as tfds

_CITATION = """
@inproceedings{49029,
title = {Wiki-40B: Multilingual Language Model Dataset},
author = {Mandy Guo and Zihang Dai and Denny Vrandecic and Rami Al-Rfou},
year = {2020},
booktitle	= {LREC 2020}
}
"""

_DESCRIPTION = """
Clean-up text for 40+ Wikipedia languages editions of pages
correspond to entities. The datasets have train/dev/test splits per language.
The dataset is cleaned up by page filtering to remove disambiguation pages,
redirect pages, deleted pages, and non-entity pages. Each example contains the
wikidata id of the entity, and the full Wikipedia article after page processing
that removes non-content sections and structured objects. The language models
trained on this corpus - including 41 monolingual models, and 2 multilingual
models - can be found at https://tfhub.dev/google/collections/wiki40b-lm/1.
"""

_LICENSE = """
This work is licensed under the Creative Commons Attribution-ShareAlike
3.0 Unported License. To view a copy of this license, visit
http://creativecommons.org/licenses/by-sa/3.0/ or send a letter to
Creative Commons, PO Box 1866, Mountain View, CA 94042, USA.
"""

_URL = "https://research.google/pubs/pub49029/"

_DATA_DIRECTORY = tfds.core.gcs_path("downloads/wiki40b/tfrecord_prod")

WIKIPEDIA_LANGUAGES = [
    "en",
    "ar",
    "zh-cn",
    "zh-tw",
    "nl",
    "fr",
    "de",
    "it",
    "ja",
    "ko",
    "pl",
    "pt",
    "ru",
    "es",
    "th",
    "tr",
    "bg",
    "ca",
    "cs",
    "da",
    "el",
    "et",
    "fa",
    "fi",
    "he",
    "hi",
    "hr",
    "hu",
    "id",
    "lt",
    "lv",
    "ms",
    "no",
    "ro",
    "sk",
    "sl",
    "sr",
    "sv",
    "tl",
    "uk",
    "vi",
]


class Wiki40bConfig(tfds.core.BuilderConfig):
  """BuilderConfig for Wiki40B."""

  def __init__(self, *, language=None, **kwargs):
    """BuilderConfig for Wiki40B.

    Args:
      language: string, the language code for the Wiki40B dataset to use.
      **kwargs: keyword arguments forwarded to super.
    """
    super(Wiki40bConfig, self).__init__(
        name=language,
        description="Wiki40B dataset for {}.".format(language),
        **kwargs,
    )
    self.language = language


_VERSION = tfds.core.Version("1.3.0")


class Wiki40b(tfds.core.BeamBasedBuilder):
  """Wiki40B: A Clean Wikipedia Dataset for Mutlilingual Language Modeling."""

  BUILDER_CONFIGS = [
      Wiki40bConfig(  # pylint:disable=g-complex-comprehension
          version=_VERSION,
          language=lang,
      )
      for lang in WIKIPEDIA_LANGUAGES
  ]

  def _info(self):
    return tfds.core.DatasetInfo(
        builder=self,
        description=_DESCRIPTION,
        features=tfds.features.FeaturesDict({
            "wikidata_id": tfds.features.Text(),
            "text": tfds.features.Text(),
            "version_id": tfds.features.Text(),
        }),
        supervised_keys=None,
        homepage=_URL,
        citation=_CITATION,
        license=_LICENSE,
    )

  def _split_generators(self, dl_manager):
    """Returns SplitGenerators."""

    del dl_manager  # Unused

    lang = self._builder_config.language

    return [
        tfds.core.SplitGenerator(
            name=tfds.Split.TRAIN,
            gen_kwargs={
                "filepaths": os.path.join(
                    _DATA_DIRECTORY, "train", "{}_examples-*".format(lang)
                )
            },
        ),
        tfds.core.SplitGenerator(
            name=tfds.Split.VALIDATION,
            gen_kwargs={
                "filepaths": os.path.join(
                    _DATA_DIRECTORY, "dev", "{}_examples-*".format(lang)
                )
            },
        ),
        tfds.core.SplitGenerator(
            name=tfds.Split.TEST,
            gen_kwargs={
                "filepaths": os.path.join(
                    _DATA_DIRECTORY, "test", "{}_examples-*".format(lang)
                )
            },
        ),
    ]

  def _build_pcollection(self, pipeline, filepaths):
    """Build PCollection of examples."""
    beam = tfds.core.lazy_imports.apache_beam
    logging.info("generating examples from = %s", filepaths)

    def _extract_content(example):
      """Extracts content from a TFExample."""
      wikidata_id = (
          example.features.feature["wikidata_id"]
          .bytes_list.value[0]
          .decode("utf-8")
      )
      text = (
          example.features.feature["text"].bytes_list.value[0].decode("utf-8")
      )
      version_id = (
          example.features.feature["version_id"]
          .bytes_list.value[0]
          .decode("utf-8")
      )

      # wikidata_id could be duplicated with different texts.
      yield wikidata_id + text, {
          "wikidata_id": wikidata_id,
          "text": text,
          "version_id": version_id,
      }

    return (
        pipeline
        | beam.io.ReadFromTFRecord(
            filepaths, coder=beam.coders.ProtoCoder(tf.train.Example)
        )
        | beam.FlatMap(_extract_content)
    )



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""locomotion dataset."""
import sys

import pytest
import tensorflow_datasets.public_api as tfds
from tensorflow_datasets.rlds.locomotion import locomotion


pytest.importorskip(
    'envlogger',
    reason=(
        '`envlogger` library might not be available for Python'
        f' {sys.version} or platform {sys.platform}; see'
        ' https://pypi.org/project/envlogger/#files'
    ),
)


class AntTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for Ant dataset."""

  DATASET_CLASS = locomotion.Locomotion
  SPLITS = {
      'train': 2,  # Number of fake train example
  }

  SKIP_TF1_GRAPH_MODE = True

  BUILDER_CONFIG_NAMES_TO_TEST = ['ant_sac_1M_single_policy_stochastic']

  DL_EXTRACT_RESULT = {'file_path': 'ant_sac_1M_single_policy_stochastic'}
  DL_DOWNLOAD_RESULT = {'file_path': 'ant_sac_1M_single_policy_stochastic'}


class Walker2dTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for Walker2d dataset."""

  DATASET_CLASS = locomotion.Locomotion
  SPLITS = {
      'train': 2,  # Number of fake train example
  }

  SKIP_TF1_GRAPH_MODE = True

  BUILDER_CONFIG_NAMES_TO_TEST = ['walker2d_sac_1M_single_policy_stochastic']

  DL_EXTRACT_RESULT = {'file_path': 'walker2d_sac_1M_single_policy_stochastic'}
  DL_DOWNLOAD_RESULT = {'file_path': 'walker2d_sac_1M_single_policy_stochastic'}


class HopperTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for Hopper dataset."""

  DATASET_CLASS = locomotion.Locomotion
  SPLITS = {
      'train': 2,  # Number of fake train example
  }

  SKIP_TF1_GRAPH_MODE = True

  BUILDER_CONFIG_NAMES_TO_TEST = ['hopper_sac_1M_single_policy_stochastic']

  DL_EXTRACT_RESULT = {'file_path': 'hopper_sac_1M_single_policy_stochastic'}
  DL_DOWNLOAD_RESULT = {'file_path': 'hopper_sac_1M_single_policy_stochastic'}


class HalfCheetahTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for HalfCheetah dataset."""

  DATASET_CLASS = locomotion.Locomotion
  SPLITS = {
      'train': 2,  # Number of fake train example
  }

  SKIP_TF1_GRAPH_MODE = True

  BUILDER_CONFIG_NAMES_TO_TEST = ['halfcheetah_sac_1M_single_policy_stochastic']

  DL_EXTRACT_RESULT = {
      'file_path': 'halfcheetah_sac_1M_single_policy_stochastic'
  }
  DL_DOWNLOAD_RESULT = {
      'file_path': 'halfcheetah_sac_1M_single_policy_stochastic'
  }


class HumanoidTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for Humanoid dataset."""

  DATASET_CLASS = locomotion.Locomotion
  SPLITS = {
      'train': 2,  # Number of fake train example
  }

  SKIP_TF1_GRAPH_MODE = True

  BUILDER_CONFIG_NAMES_TO_TEST = ['humanoid_sac_15M_single_policy_stochastic']

  DL_EXTRACT_RESULT = {'file_path': 'humanoid_sac_15M_single_policy_stochastic'}
  DL_DOWNLOAD_RESULT = {
      'file_path': 'humanoid_sac_15M_single_policy_stochastic'
  }


if __name__ == '__main__':
  tfds.testing.test_main()



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Structured datasets."""

from tensorflow_datasets.structured.amazon_us_reviews import AmazonUSReviews
from tensorflow_datasets.structured.ble_wind_field import BLEWindField
from tensorflow_datasets.structured.cherry_blossoms import CherryBlossoms
from tensorflow_datasets.structured.covid19 import Covid19
from tensorflow_datasets.structured.cs_restaurants import CSRestaurants
from tensorflow_datasets.structured.dart import Dart
from tensorflow_datasets.structured.diamonds import Diamonds
from tensorflow_datasets.structured.e2e_cleaned import E2eCleaned
from tensorflow_datasets.structured.efron_morris_75 import EfronMorris75
from tensorflow_datasets.structured.forest_fires import ForestFires
from tensorflow_datasets.structured.genomics_ood import GenomicsOod
from tensorflow_datasets.structured.german_credit_numeric import GermanCreditNumeric
from tensorflow_datasets.structured.higgs import Higgs
from tensorflow_datasets.structured.howell import Howell
from tensorflow_datasets.structured.iris import Iris
from tensorflow_datasets.structured.kddcup99 import Kddcup99
from tensorflow_datasets.structured.movielens import Movielens
from tensorflow_datasets.structured.penguins import Penguins
from tensorflow_datasets.structured.proteinnet import ProteinNet
from tensorflow_datasets.structured.radon import Radon
from tensorflow_datasets.structured.rock_you import RockYou
from tensorflow_datasets.structured.titanic import Titanic
from tensorflow_datasets.structured.web_graph import WebGraph
from tensorflow_datasets.structured.web_nlg.web_nlg import WebNlg
from tensorflow_datasets.structured.wiki_bio import WikiBio
from tensorflow_datasets.structured.wiki_table_questions import WikiTableQuestions
from tensorflow_datasets.structured.wiki_table_text import WikiTableText
from tensorflow_datasets.structured.wine_quality import WineQuality



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""CLI for Tensorflow Datasets.

TFDS CLI to help creates and build datasets (e.g. `tfds new my_dataset`,
`tfds build`,...)

See: https://www.tensorflow.org/datasets/cli

"""

import argparse
import logging as python_logging
import sys
from typing import List

from absl import app
from absl import flags
from absl import logging
from absl.flags import argparse_flags

import tensorflow_datasets.public_api as tfds

# Import commands
from tensorflow_datasets.scripts.cli import build
from tensorflow_datasets.scripts.cli import new
from tensorflow_datasets.scripts.utils import flag_utils

FLAGS = flags.FLAGS


def _parse_flags(argv: List[str]) -> argparse.Namespace:
  """Command lines flag parsing."""
  argv = flag_utils.normalize_flags(argv)  # See b/174043007 for context.

  parser = argparse_flags.ArgumentParser(
      description='Tensorflow Datasets CLI tool',
      allow_abbrev=False,
  )
  parser.add_argument(
      '--version',
      action='version',
      version='TensorFlow Datasets: ' + tfds.__version__,
  )
  parser.set_defaults(subparser_fn=lambda _: parser.print_help())
  # Register sub-commands
  subparser = parser.add_subparsers(title='command')
  build.register_subparser(subparser)
  new.register_subparser(subparser)
  return parser.parse_args(argv[1:])


def main(args: argparse.Namespace) -> None:

  # From the CLI, all datasets are visible
  tfds.core.visibility.set_availables(
      [
          tfds.core.visibility.DatasetType.TFDS_PUBLIC,
          tfds.core.visibility.DatasetType.COMMUNITY_PUBLIC,
      ]
  )

  # By default, ABSL won't display any `logging.info` unless the
  # user explicitly set `--logtostderr`.
  # For usability, we activate by default the python log information, but
  # not the C++ ones (which are too verbose).
  # `logtostderr` may not be defined if `main()` is called directly without
  # `absl.run` (e.g. open source `pytest` tests)
  if not FLAGS.is_parsed() or (
      # If user explicitly request logs, keep C++ logger
      not FLAGS.logtostderr
      and not FLAGS.alsologtostderr
  ):
    # Using cleaner, less verbose logger
    formatter = python_logging.Formatter(
        '{levelname}[{filename}]: {message}', style='{'
    )
    logging.use_python_logging(quiet=True)
    logging.set_verbosity(logging.INFO)
    python_handler = logging.get_absl_handler().python_handler
    python_handler.setFormatter(formatter)
    # Replace `sys.stderr` by the TQDM file
    new_stream = tfds.core.utils.tqdm_utils.TqdmStream()
    if sys.version_info >= (3, 7):
      python_handler.setStream(new_stream)
    else:
      python_handler.stream.flush()
      python_handler.stream = new_stream

  # Launch the subcommand defined in the subparser (or default to print help)
  args.subparser_fn(args)


def launch_cli() -> None:
  """Parse arguments and launch the CLI main function."""
  app.run(main, flags_parser=_parse_flags)


if __name__ == '__main__':
  # Entry-points in `setup.py` launch the `launch_cli()` function directly, so
  # the code in `if __name__ == '__main__'` is never executed.
  launch_cli()



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""rlu_atari_checkpoints_ordered dataset."""

from __future__ import annotations

from typing import Any, Dict

from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
import tensorflow_datasets.public_api as tfds
from tensorflow_datasets.rl_unplugged import atari_utils
from tensorflow_datasets.rl_unplugged import rlu_common

_EXTRA_DESCRIPTION = """
  Each of the configurations is broken into splits. Splits correspond to
  checkpoints of 1M steps (note that the number of episodes may difer).
  Checkpoints are ordered in time (so checkpoint 0 ran before checkpoint 1).

  Episodes within each split are ordered. Check
  https://www.tensorflow.org/datasets/determinism if you want to ensure
  that you read episodes in order.

  This dataset corresponds to the one used in the DQN replay paper.
  https://research.google/tools/datasets/dqn-replay/
"""


class RluAtariCheckpointsOrdered(rlu_common.RLUBuilder):
  """DatasetBuilder for RLU Atari with one split per checkpoint."""

  _SHARDS = 50
  _SPLITS = 50
  _INPUT_FILE_PREFIX = 'gs://rl_unplugged/atari_episodes_ordered'

  VERSION = tfds.core.Version('1.1.0')
  RELEASE_NOTES = {
      '1.0.0': 'Initial release.',
      '1.1.0': 'Removed redundant clipped reward fields.',
  }

  BUILDER_CONFIGS = atari_utils.builder_configs()

  def _info(self) -> tfds.core.DatasetInfo:
    """Returns the dataset metadata with shuffling disabled."""
    # pylint: disable=protected-access
    return tfds.core.DatasetInfo(
        builder=self,
        description=rlu_common._DESCRIPTION + self.get_description(),
        features=self.get_features_dict(),
        supervised_keys=None,
        homepage=rlu_common._HOMEPAGE,
        citation=self.get_citation(),
        disable_shuffling=True,
    )
    # pylint: enable=protected-access

  def get_features_dict(self):
    return atari_utils.features_dict()

  def get_description(self):
    return atari_utils.description() + _EXTRA_DESCRIPTION

  def get_citation(self):
    return atari_utils.citation()

  def get_file_prefix(self):
    run = self.builder_config.run
    game = self.builder_config.game
    return atari_utils.file_prefix(self._INPUT_FILE_PREFIX, run, game)

  def num_shards(self):
    return atari_utils.num_shards(self.builder_config.game, self._SHARDS)

  def get_episode_id(self, episode):
    # TODO(b/209933106): Ordered datasets need an int id so we cannot use the
    # one from atari_utils. We estimate that there are at most 20k episodes
    # per checkpoint (numbers star at 0), 1e10 should be a more than safe
    # number.
    episode_id = episode['checkpoint_id'] * 1e10 + episode['episode_id']
    return int(episode_id)

  def tf_example_to_step_ds(
      self, tf_example: tf.train.Example
  ) -> Dict[str, Any]:
    return atari_utils.atari_example_to_rlds(tf_example)

  def get_splits(self):
    checkpoints = {}
    for i in range(self.num_shards()):
      paths = {
          'file_paths': rlu_common.filename(
              prefix=self.get_file_prefix(),
              num_shards=self.num_shards(),
              shard_id=i,
          )
      }
      checkpoints[f'checkpoint_{i:02d}'] = self._generate_examples(paths)
    return checkpoints

  def _generate_examples(self, paths):
    """Yields examples."""

    file_path = paths['file_paths']

    return self.generate_examples_one_file(file_path)



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""controlled_noisy_web_labels dataset."""

from tensorflow_datasets.image_classification.controlled_noisy_web_labels import controlled_noisy_web_labels
import tensorflow_datasets.public_api as tfds


class ControlledNoisyWebLabelsTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for controlled_noisy_web_labels dataset."""

  DATASET_CLASS = controlled_noisy_web_labels.ControlledNoisyWebLabels
  SPLITS = {
      'train_00': 3,  # Number of fake train examples
      'train_05': 6,  # Number of fake train examples
      'train_10': 6,  # Number of fake train examples
      'train_15': 6,  # Number of fake train examples
      'train_20': 6,  # Number of fake train examples
      'train_30': 6,  # Number of fake train examples
      'train_40': 6,  # Number of fake train examples
      'train_50': 6,  # Number of fake train examples
      'train_60': 6,  # Number of fake train examples
      'train_80': 6,  # Number of fake train examples
      'validation': 3,  # Number of fake test examples
  }

  DL_EXTRACT_RESULT = {
      'mini_train': 'train.csv',
      'mini_val': 'val.csv',
      'mini_test': 'test.csv',
  }

  OVERLAPPING_SPLITS = [
      'train_00',
      'train_05',
      'train_10',
      'train_15',
      'train_20',
      'train_30',
      'train_40',
      'train_50',
      'train_60',
      'train_80',
  ]

  SKIP_CHECKSUMS = True


if __name__ == '__main__':
  tfds.testing.test_main()
